# -*- coding: utf-8 -*-

import datetime

from django.db import migrations
from django.utils.timezone import utc

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="article",
            name="last_modified",
            field=touchtechnology.common.db.models.DateTimeField(
                default=datetime.datetime(2015, 3, 23, 11, 10, 25, 290897, tzinfo=utc),
                auto_now=True,
            ),
            preserve_default=False,
        ),
    ]



from django import forms

from touchtechnology.common.forms.fields import SelectDateTimeField


class TestForm1(forms.Form):
    """
    The `timestamp` field is declared inside of the constructor because if we
    do not the field instance will be evaluated at module import time. In a
    real site this would be fine, however in our test cases we need to be able
    to override the USE_TZ setting to test for both modes of operation.
    """

    def __init__(self, *args, **kwargs):
        super(TestForm1, self).__init__(*args, **kwargs)
        self.fields["timestamp"] = SelectDateTimeField()



# -*- coding: utf-8 -*-


import logging
from importlib import import_module

from django.contrib.contenttypes.fields import GenericRelation
from django.core.cache import caches
from django.db import models
from django.db.models import DateTimeField
from django.utils.module_loading import import_string
from django.utils.text import slugify
from django.utils.translation import gettext, gettext_lazy as _
from first import first

from touchtechnology.admin.mixins import AdminUrlMixin
from touchtechnology.common.db.models import (
    BooleanField,
    HTMLField,
    TemplatePathField,
)
from touchtechnology.content.app_settings import (
    NODE_CACHE,
    PAGE_CONTENT_CLASSES,
    PAGE_TEMPLATE_BASE,
    PAGE_TEMPLATE_FOLDER,
    PAGE_TEMPLATE_REGEX,
)

logger = logging.getLogger(__name__)

PAGE_CONTENT_CLASS_CHOICES = zip(PAGE_CONTENT_CLASSES, PAGE_CONTENT_CLASSES)

PAGE_CONTENT_CLASS_DEFAULT = first(PAGE_CONTENT_CLASSES)

SITE_CACHE_KEY = "_site_cache"


class AdminUrlModel(AdminUrlMixin, models.Model):
    class Meta:
        abstract = True

    def _get_admin_namespace(self):
        return "admin:content:%s" % self._meta.model_name

    def _get_url_args(self):
        return (self.id,)

    def __repr__(self):
        return "<{}: {}>".format(self.__class__, self)


class Page(models.Model):

    # allows us to customise the template on a per page basis

    template = TemplatePathField(
        max_length=200,
        blank=True,
        template_base=PAGE_TEMPLATE_BASE,
        template_folder=PAGE_TEMPLATE_FOLDER,
        match=PAGE_TEMPLATE_REGEX,
        recursive=True,
        verbose_name=_("Template"),
    )

    # SEO metadata fields

    keywords = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Keywords"),
        help_text=_(
            "This should be a comma-separated list of terms that "
            "indicate the content of this page - used to assist "
            "Search Engines rank your page."
        ),
    )

    description = models.CharField(
        max_length=255,
        blank=True,
        verbose_name=_("Description"),
        help_text=_(
            "Search Engines may use this when determining the "
            "relevance of your page."
        ),
    )

    last_modified = DateTimeField(auto_now=True)

    def __repr__(self):
        return "<Page: #{}>".format(self.pk)


class Content(models.Model):

    # FIXME and put the ``copy`` attribute on the ``PageContent`` model, we
    # aren't actually reusing the ``Content`` in any way, so simplify the
    # implementation to move the data back up and onto.

    # The M2M relation on ``Page`` will become a FK on ``PageContent``
    # instead.

    copy = HTMLField(blank=True)

    def __repr__(self):
        return "<Content: #%d>" % self.pk


class PageContent(models.Model):

    page = models.ForeignKey(
        Page, related_name="content", verbose_name=_("Page"), on_delete=models.PROTECT
    )
    label = models.SlugField(
        max_length=20,
        choices=PAGE_CONTENT_CLASS_CHOICES,
        default=PAGE_CONTENT_CLASS_DEFAULT,
        verbose_name=_("CSS class"),
    )
    sequence = models.PositiveIntegerField(verbose_name=_("Sequence"))
    copy = HTMLField(blank=True, verbose_name=_("Copy"))

    last_modified = DateTimeField(auto_now=True)

    class Meta:
        ordering = ("sequence",)

    def __repr__(self):
        return "<PageContent: #%d, page=%d>" % (self.pk, self.page.pk)


class Chunk(AdminUrlModel):

    slug = models.SlugField(verbose_name=_("Slug"))
    copy = HTMLField(blank=True, verbose_name=_("Page Copy"))

    def __str__(self):
        return self.slug


class NodeContent(models.Model):

    node = models.ForeignKey(
        "common.SitemapNode",
        related_name="contents",
        verbose_name=_("Node"),
        on_delete=models.PROTECT,
    )
    copy = HTMLField(blank=True, verbose_name=_("Page Copy"))


class Placeholder(models.Model):

    path = models.CharField(max_length=255, verbose_name=_("Module path"))
    namespace = models.CharField(
        max_length=255, db_index=True, verbose_name=_("Namespace")
    )

    nodes = GenericRelation("common.SitemapNode", verbose_name=_("Nodes"))

    class Meta:
        ordering = ("path",)

    def __str__(self):
        try:
            cls = import_string(self.path)
        except (ImportError, ValueError):
            return gettext("Not installed")
        return cls.verbose_name()

    def _install(self, parent=None):
        """
        Adding an internal method to allow us to rig up all the applications.
        """
        if not self.nodes.count():
            module_path, site = self.path.rsplit(".", 1)
            module = import_module(module_path)
            if hasattr(module, site):
                instance = getattr(module, site)
                self.nodes.create(
                    title=instance.name.title(),
                    slug=slugify(instance.name),
                    parent=parent,
                )

    @property
    def module(self):
        if not hasattr(self, "_module"):
            module_path = self.path.split(".")
            while module_path:
                try:
                    module = import_module(".".join(module_path))
                    if hasattr(module, "VERSION"):
                        self._module = module
                        break
                except ImportError:
                    module = None
                module_path.pop()
        return self._module

    def callable(self):
        try:
            return import_string(self.path)
        except ValueError:
            return None

    def site(self, node):
        cache_key = "node%d" % node.pk
        cls = import_string(self.path)
        site = cls(node=node, name=cache_key, **node.kwargs)
        return site

    def invalidate_cache(self):
        # Since this is forced cache invalidation, we should log it.
        logger.info("Forced invalidation of Application cache.")
        cache = caches[NODE_CACHE]
        return cache.clear()

    invalidate_cache.alters_data = True


class Redirect(AdminUrlModel):

    source_url = models.CharField(
        max_length=250,
        verbose_name=_("Source URL"),
        help_text=_("The path that will trigger the redirection."),
    )

    destination_url = models.CharField(
        max_length=500,
        verbose_name=_("Destination URL"),
        help_text=_("The URL or path that the browser will be sent to."),
    )

    label = models.CharField(max_length=100, blank=True, verbose_name=_("Label"))

    active = BooleanField(default=True, verbose_name=_("Active"))
    permanent = BooleanField(default=False, verbose_name=_("Permanent"))

    class Meta:
        ordering = ("destination_url",)

    def __str__(self):
        if self.label:
            return self.label
        return "{src} → {dst}".format(src=self.source_url, dst=self.destination_url)



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("competition", "0047_season_live_stream_oauth2"),
    ]

    operations = [
        migrations.AddField(
            model_name="season",
            name="live_stream_thumbnail",
            field=models.URLField(null=True),
        ),
    ]



# -*- coding: utf-8 -*-

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0012_hashtag_include_symbol"),
    ]

    operations = [
        migrations.AlterField(
            model_name="simplescorematchstatistic",
            name="played",
            field=models.SmallIntegerField(
                default=0,
                blank=True,
                validators=[
                    django.core.validators.MinValueValidator(0),
                    django.core.validators.MaxValueValidator(1),
                ],
            ),
        ),
    ]



from django.contrib.sitemaps.views import sitemap
from django.urls import path

from touchtechnology.admin.sites import site
from touchtechnology.common.sitemaps import NodeSitemap
from touchtechnology.common.sites import AccountsSite

accounts = AccountsSite()

urlpatterns = [
    path("admin/", site.urls),
    path("accounts/", accounts.urls),
    path("sitemap.xml", sitemap, {"sitemaps": {"nodes": NodeSitemap}}, name="sitemap"),
]



# -*- coding: utf-8 -*-

from django.db import migrations


def link_person_model(apps, schema_editor):
    Club = apps.get_model("competition", "Club")
    Person = apps.get_model("competition", "Person")

    # Update the Club.primary_uuid field, with no reverse relation we need to
    # edit it on the object directly.
    for club in Club.objects.exclude(primary__isnull=True):
        club.primary_uuid = club.primary
        club.save()

    # Update all the reverse relations per person.
    for person in Person.objects.all():
        person.clubassociation_set.update(person_uuid=person.uuid)
        person.teamassociation_set.update(person_uuid=person.uuid)
        person.seasonassociation_set.update(person_uuid=person.uuid)
        person.statistics.update(player_uuid=person.uuid)


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0020_competition__add__person_uuid"),
    ]

    operations = [
        migrations.RunPython(link_person_model, reverse_code=migrations.RunPython.noop),
    ]



from django.db import models

from touchtechnology.common.db.models import DateTimeField


class TestDateTimeField(models.Model):

    datetime = DateTimeField(blank=True, null=True)

    class Meta:
        ordering = ("datetime",)

    def __str__(self):
        return str(self.datetime)


class Relative(models.Model):

    name = models.CharField(max_length=10, blank=True, null=True)
    link = models.ForeignKey("TestDateTimeField", on_delete=models.CASCADE)



import zoneinfo

from django import forms
from django.utils import timezone


def timezone_choice(tzname, country=None):
    """
    Given a timezone string, convert it to a two-tuple suitable
    for a choice in a FormField.

    >>> timezone_choice('UTC')
    ('UTC', 'UTC')
    >>> timezone_choice('Australia/Sydney')
    ('Australia/Sydney', 'Sydney')
    >>> timezone_choice('America/Indiana/Indianapolis')
    ('America/Indiana/Indianapolis', 'Indianapolis (Indiana)')
    """
    try:
        base, rest = tzname.split("/", 1)
    except ValueError:
        base, rest = None, tzname
    rest = rest.replace("_", " ")
    try:
        mid, rest = rest.split("/")
        if mid != base and mid != country:
            rest = "{0} ({1})".format(rest, mid)
    except ValueError:
        pass
    return (tzname, rest)


MONTH_CHOICES = (
    ("", ""),
    (1, "January"),
    (2, "February"),
    (3, "March"),
    (4, "April"),
    (5, "May"),
    (6, "June"),
    (7, "July"),
    (8, "August"),
    (9, "September"),
    (10, "October"),
    (11, "November"),
    (12, "December"),
)

DAY_CHOICES = [("", "")] + [(day, day) for day in range(1, 32)]
HOUR_CHOICES = [("", "")] + [(h, h) for h in range(24)]
MINUTE_CHOICES = [("", "")] + [(m, f"{m:02d}") for m in range(60)]
TIMEZONE_CHOICES = [
    timezone_choice(tz, tz.split("/", 1)[0])
    for tz in sorted(zoneinfo.available_timezones())
]


class SelectTimezoneForm(forms.Form):
    """
    Used by touchtechnology.common.context_processors.tz to generate a
    dropdown list of timezones to be used with the TimeZone middleware and
    set-timezone view.
    """

    timezone = forms.ChoiceField(
        choices=TIMEZONE_CHOICES, initial=timezone.get_current_timezone
    )



import datetime

import factory
import factory.fuzzy
from example_app import models
from factory.django import DjangoModelFactory


class TestDateTimeFieldFactory(DjangoModelFactory):
    class Meta:
        model = models.TestDateTimeField

    datetime = factory.fuzzy.FuzzyDateTime(
        datetime.datetime(2013, 7, 15, tzinfo=datetime.timezone.utc),
        datetime.datetime(2015, 6, 15, tzinfo=datetime.timezone.utc),
        force_microsecond=0,
    )


class RelativeFactory(DjangoModelFactory):
    class Meta:
        model = models.Relative

    name = factory.Sequence(lambda n: "Sample %d" % n)

    link = factory.SubFactory(TestDateTimeFieldFactory)



from django.test.utils import override_settings
from test_plus import TestCase


@override_settings(ROOT_URLCONF="example_app.urls")
class DateTimeFieldTest(TestCase):
    def test_validate_form_output_tz(self):
        with self.settings(USE_TZ=False):
            self.get("datetime:index")
            self.assertResponseNotContains(
                '<option value="Australia/Sydney">Sydney</option>'
            )

        with self.settings(USE_TZ=True):
            self.get("datetime:index")
            self.assertResponseContains(
                '<option value="Australia/Sydney">Sydney</option>'
            )

    def test_submit_date_time_valid(self):
        data = {
            "form-datetime_0": "1",
            "form-datetime_1": "1",
            "form-datetime_2": "2014",
            "form-datetime_3": "0",
            "form-datetime_4": "0",
            "formset-TOTAL_FORMS": "0",
            "formset-INITIAL_FORMS": "0",
            "formset-MAX_NUM_FORMS": "1000",
        }

        with self.settings(USE_TZ=False):
            self.post("datetime:index", data=data)
            self.response_302()

        data.update(
            {
                "form-datetime_5": "Australia/Sydney",
            }
        )

        with self.settings(USE_TZ=True):
            self.post("datetime:index", data=data)
            self.response_302()

    def test_submit_date_time_blank(self):
        data = {
            "form-datetime_0": "",
            "form-datetime_1": "",
            "form-datetime_2": "",
            "form-datetime_3": "",
            "form-datetime_4": "",
            "formset-TOTAL_FORMS": "0",
            "formset-INITIAL_FORMS": "0",
            "formset-MAX_NUM_FORMS": "1000",
        }

        with self.settings(USE_TZ=False):
            self.post("datetime:index", data=data)
            self.response_302()

        data.update(
            {
                "form-datetime_5": "",
            }
        )

        with self.settings(USE_TZ=True):
            self.post("datetime:index", data=data)
            self.response_302()



# -*- coding: utf-8 -*-

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0009_remove_byeteam"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="matchvideo",
            options={"verbose_name": "video"},
        ),
        migrations.RemoveField(
            model_name="matchvideo",
            name="approved_by",
        ),
        migrations.AlterField(
            model_name="matchvideo",
            name="url",
            field=models.URLField(max_length=255, verbose_name="URL"),
            preserve_default=True,
        ),
    ]



from django.test.utils import override_settings
from test_plus import TestCase

from tournamentcontrol.competition.models import LadderSummary
from tournamentcontrol.competition.tests import factories

ORIGINAL_BEHAVIOUR = "tournamentcontrol.competition.rank.points_func"
CORRECT_BEHAVIOUR = "tournamentcontrol.competition.rank.correct_points_func"


class RankingLadderEntryTestCase(TestCase):
    """
    The default implementation awards the following bonus multipliers:

    -   double the win points for wins > 15
    -   win points for wins > 10
    -   half win points for wins > 5
    -   half loss points for losses < 2

    The DYNAMIC_FIELDS are computed with annotations in the LadderEntryQuerySet
    using the function addressed by the `TOURNAMENTCONTROL_RANK_POINTS_FUNC`
    setting.

    Our original implementation had a bug where margin bonuses for winning
    would be accumulated if more than one condition was satisfied. We test both
    the original and desired behaviour.
    """

    DYNAMIC_FIELDS = ("margin", "diff", "rank_points")

    def assertLadderEntries(
        self, home_score, away_score, expected, expected_summaries=2, **kwargs
    ):
        match = factories.MatchFactory.create(
            home_team_score=home_score, away_team_score=away_score, **kwargs
        )
        self.assertCountEqual(
            match.ladder_entries.values_list(*self.DYNAMIC_FIELDS), expected
        )
        self.assertEqual(
            expected_summaries, LadderSummary.objects.filter(stage=match.stage).count()
        )


@override_settings(TOURNAMENTCONTROL_RANK_POINTS_FUNC=ORIGINAL_BEHAVIOUR)
class OriginalBehaviourTests(RankingLadderEntryTestCase):
    def test_bug_win_no_margin_bonus(self):
        "win by <=5, loss by >=2"
        self.assertLadderEntries(3, 1, [(2, 2, 15), (2, -2, 5)])

    def test_bug_win_margin_bonus_small(self):
        "win by >5"
        self.assertLadderEntries(7, 1, [(6, 6, 22.5), (6, -6, 5)])

    def test_bug_win_margin_bonus_medium(self):
        "win by >10"
        self.assertLadderEntries(12, 1, [(11, 11, 37.5), (11, -11, 5)])

    def test_bug_win_margin_bonus_large(self):
        "win by >15"
        self.assertLadderEntries(17, 1, [(16, 16, 67.5), (16, -16, 5)])

    def test_bug_loss_margin_bonus(self):
        "loss by <2"
        self.assertLadderEntries(2, 1, [(1, 1, 15), (1, -1, 7.5)])


@override_settings(TOURNAMENTCONTROL_RANK_POINTS_FUNC=CORRECT_BEHAVIOUR)
class CorrectBehaviourTests(RankingLadderEntryTestCase):
    def test_win_no_margin_bonus(self):
        "win by <=5, loss by >=2"
        self.assertLadderEntries(3, 1, [(2, 2, 15), (2, -2, 5)])

    def test_win_margin_bonus_small(self):
        "win by >5"
        self.assertLadderEntries(7, 1, [(6, 6, 22.5), (6, -6, 5)])

    def test_win_margin_bonus_medium(self):
        "win by >10"
        self.assertLadderEntries(12, 1, [(11, 11, 30), (11, -11, 5)])

    def test_win_margin_bonus_large(self):
        "win by >15"
        self.assertLadderEntries(17, 1, [(16, 16, 45), (16, -16, 5)])

    def test_loss_margin_bonus(self):
        "loss by <2"
        self.assertLadderEntries(2, 1, [(1, 1, 15), (1, -1, 7.5)])


@override_settings(TOURNAMENTCONTROL_RANK_POINTS_FUNC=CORRECT_BEHAVIOUR)
class Issue27Tests(RankingLadderEntryTestCase):
    def test_match_not_included_in_ladder(self):
        self.assertLadderEntries(
            2, 1, [(1, 1, 15), (1, -1, 7.5)], include_in_ladder=False
        )
        # Extra test - we expect the LadderSummary to exist because the Stage will keep
        # a ladder, but this particular match is excluded from it.
        self.assertCountEqual(
            LadderSummary.objects.values_list("played", "score_for", "score_against"),
            [(0, 0, 0), (0, 0, 0)],
        )

    def test_stage_does_not_keep_ladder(self):
        self.assertLadderEntries(
            2,
            1,
            [(1, 1, 15), (1, -1, 7.5)],
            expected_summaries=0,
            stage__keep_ladder=False,
        )

    def test_neither_included_in_ladder_or_stage_keep_ladder(self):
        self.assertLadderEntries(
            2,
            1,
            [(1, 1, 15), (1, -1, 7.5)],
            expected_summaries=0,
            include_in_ladder=False,
            stage__keep_ladder=False,
        )



from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("common", "0006_model_definition_sync"),
    ]

    operations = [
        migrations.AlterField(
            model_name="sitemapnode",
            name="kwargs",
            field=models.JSONField(default=dict),
        ),
        migrations.AlterField(
            model_name="sitemapnode",
            name="level",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="sitemapnode",
            name="lft",
            field=models.PositiveIntegerField(editable=False),
        ),
        migrations.AlterField(
            model_name="sitemapnode",
            name="rght",
            field=models.PositiveIntegerField(editable=False),
        ),
    ]



from rest_framework import serializers
from rest_framework_nested.serializers import NestedHyperlinkedModelSerializer

from tournamentcontrol.competition import models

from .viewsets import SlugViewSet


class PlaceSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Place
        fields = ("id", "title", "abbreviation", "timezone")


class RefereeClubSerializer(serializers.ModelSerializer):
    class Meta:
        model = models.Club
        fields = ("title", "abbreviation")


class ListRefereeSerializer(serializers.ModelSerializer):
    uuid = serializers.ReadOnlyField(source="person.uuid")
    first_name = serializers.ReadOnlyField(source="person.first_name")
    last_name = serializers.ReadOnlyField(source="person.last_name")

    club = RefereeClubSerializer(read_only=True)

    class Meta:
        model = models.SeasonReferee
        fields = ("id", "uuid", "first_name", "last_name", "club")


class ListDivisionSerializer(NestedHyperlinkedModelSerializer):
    parent_lookup_kwargs = {
        "competition_slug": "season__competition__slug",
        "season_slug": "season__slug",
    }

    class Meta:
        model = models.Division
        fields = ("title", "slug", "url")
        extra_kwargs = {
            "url": {"lookup_field": "slug"},
            "division": {"lookup_field": "slug"},
        }


class ListSeasonSerializer(NestedHyperlinkedModelSerializer):
    parent_lookup_kwargs = {"competition_slug": "competition__slug"}

    class Meta:
        model = models.Season
        fields = ("title", "slug", "url")
        extra_kwargs = {
            "url": {"lookup_field": "slug"},
            "competition": {"lookup_field": "slug"},
        }


class SeasonSerializer(ListSeasonSerializer):
    divisions = ListDivisionSerializer(many=True, read_only=True)
    referees = ListRefereeSerializer(many=True, read_only=True)

    class Meta(ListSeasonSerializer.Meta):
        fields = ("title", "slug", "url", "divisions", "referees")


class SeasonViewSet(SlugViewSet):
    serializer_class = SeasonSerializer
    list_serializer_class = ListSeasonSerializer

    def get_queryset(self):
        return models.Season.objects.filter(
            competition__slug=self.kwargs["competition_slug"]
        )



import unittest
from datetime import date, datetime, time
from zoneinfo import ZoneInfo

from django import VERSION
from django.template import Context, Template
from django.urls import reverse
from test_plus import TestCase as BaseTestCase

from touchtechnology.common.tests.factories import UserFactory
from tournamentcontrol.competition.models import Team
from tournamentcontrol.competition.tests import factories
from tournamentcontrol.competition.utils import round_robin


class TestCase(BaseTestCase):
    def setUp(self):
        super(TestCase, self).setUp()
        self.superuser = UserFactory.create(is_staff=True, is_superuser=True)

    def assertGoodEditView(self, viewname, *args, **kwargs):
        test_query_count = kwargs.pop("test_query_count", 50)
        self.assertLoginRequired(viewname, *args)
        with self.login(self.superuser):
            self.assertGoodView(viewname, test_query_count=test_query_count, *args)
            if kwargs:
                self.post(viewname, *args, **kwargs)
                self.response_302()

    def assertGoodDeleteView(self, viewname, *args):
        self.assertLoginRequired(viewname, *args)
        with self.login(self.superuser):
            self.get(viewname, *args)
            self.response_405()
            self.post(viewname, *args)
            self.response_302()

    def assertGoodNamespace(self, instance, **kwargs):
        namespace = instance._get_admin_namespace()
        args = instance._get_url_args()
        self.assertGoodEditView("%s:add" % namespace, *args[:-1], **kwargs)
        self.assertGoodEditView("%s:edit" % namespace, *args, **kwargs)
        self.assertGoodDeleteView("%s:delete" % namespace, *args)


class TemplateTests(TestCase):
    @classmethod
    def setUpClass(cls):
        super(TemplateTests, cls).setUpClass()
        stage1 = factories.StageFactory.create()
        stage2 = factories.StageFactory.create(division=stage1.division, follows=stage1)
        ground = factories.GroundFactory.create(venue__season=stage2.division.season)
        cls.team = factories.TeamFactory.create(
            title="Chees\u00e9\u00a0&\u00a0Crackers", division=stage1.division
        )
        factories.MatchFactory.create(
            home_team=None,
            away_team=None,
            home_team_eval="P1",
            away_team_eval="P2",
            stage=stage2,
            date=date(2017, 2, 13),
            time=time(9, 0),
            play_at=ground,
            datetime=datetime(2017, 2, 13, 9, tzinfo=ZoneInfo("UTC")),
        )
        factories.MatchFactory.create(
            home_team=cls.team,
            away_team=None,
            home_team_eval="P3",
            away_team_eval="P4",
            stage=stage2,
            date=date(2017, 2, 13),
            time=time(10, 0),
            play_at=ground,
            datetime=datetime(2017, 2, 13, 10, tzinfo=ZoneInfo("UTC")),
        )
        cls.stage = stage2
        cls.division = stage2.division
        cls.season = cls.division.season
        cls.competition = cls.season.competition

    @classmethod
    def tearDownClass(cls):
        super(TemplateTests, cls).tearDownClass()

    def test_team_title(self):
        template = Template("{% load common %}{{ team.title|htmlentities }}")
        context = Context({"team": self.team})
        output = template.render(context)
        self.assertEqual(output, "Chees&eacute;&nbsp;&amp;&nbsp;Crackers")

    def test_match_runsheet(self):
        self.assertLoginRequired(
            "admin:fixja:match-runsheet",
            self.competition.pk,
            self.season.pk,
            "20170213",
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:match-runsheet",
                self.competition.pk,
                self.season.pk,
                "20170213",
            )
            self.assertResponseContains('<td class="team">1st</td>')
            self.assertResponseContains(
                '<td class="team">2nd </td>'
            )  # whitespace deliberate
            self.assertResponseContains(
                '<td class="team">Chees&eacute;&nbsp;&amp;&nbsp;Crackers</td>'
            )

    def test_match_grid(self):
        self.assertLoginRequired(
            "admin:fixja:match-grid",
            self.competition.pk,
            self.season.pk,
            "20170213",
            "html",
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:match-grid",
                self.competition.pk,
                self.season.pk,
                "20170213",
                "html",
            )
            self.assertResponseContains(
                "<h4>%s<br /><small>%s</small></h4>" % (self.division, self.stage)
            )
            self.assertResponseContains(
                "<p>1st <br>2nd</p>"
            )  # whitespace & br style deliberate
            self.assertResponseContains(
                "<p>Chees&eacute;&nbsp;&amp;&nbsp;Crackers<br>4th</p>"
            )


class GoodViewTests(TestCase):
    def test_reorder_down(self):
        stage = factories.StageFactory.create()
        pool_a, pool_b, pool_c = factories.StageGroupFactory.create_batch(
            stage=stage, size=3
        )

        with self.login(self.superuser):
            # Without a HTTP_REFERER set this should throw a 404
            self.get("admin:fixja:reorder", "stagegroup", pool_b.pk, "down")
            self.response_404()

            # The first time we try to shift this down it will work because it
            # isn't last in it's set.
            self.get(
                "admin:fixja:reorder",
                "stagegroup",
                pool_b.pk,
                "down",
                extra=dict(HTTP_REFERER="http://testserver/"),
            )
            self.response_302()

            # This time, however, it should fail because we're now last.
            self.get(
                "admin:fixja:reorder",
                "stagegroup",
                pool_b.pk,
                "down",
                extra=dict(HTTP_REFERER="http://testserver/"),
            )
            self.response_404()

    def test_reorder_up(self):
        stage = factories.StageFactory.create()
        pool_a, pool_b, pool_c = factories.StageGroupFactory.create_batch(
            stage=stage, size=3
        )

        with self.login(self.superuser):
            # Without a HTTP_REFERER set this should throw a 404
            self.get("admin:fixja:reorder", "stagegroup", pool_b.pk, "up")
            self.response_404()

            # The first time we try to shift this up it will work because it
            # isn't first in it's set.
            self.get(
                "admin:fixja:reorder",
                "stagegroup",
                pool_b.pk,
                "up",
                extra=dict(HTTP_REFERER="http://testserver/"),
            )
            self.response_302()

            # This time, however, it should fail because we're now first.
            self.get(
                "admin:fixja:reorder",
                "stagegroup",
                pool_b.pk,
                "up",
                extra=dict(HTTP_REFERER="http://testserver/"),
            )
            self.response_404()

    def test_reorder_left(self):
        stage = factories.StageFactory.create()
        pool_a, pool_b, pool_c = factories.StageGroupFactory.create_batch(
            stage=stage, size=3
        )

        with self.login(self.superuser):
            # A direction of "left" is invalid and should throw a 404
            self.get(
                "admin:fixja:reorder",
                pool_b.pk,
                "left",
                extra=dict(HTTP_REFERER="http://testserver/"),
            )
            self.response_404()

    def test_index(self):
        self.assertLoginRequired("admin:fixja:index")
        with self.login(self.superuser):
            self.get("admin:fixja:index")
            self.response_302()

    def test_competition(self):
        competition = factories.CompetitionFactory.create()
        self.assertLoginRequired("admin:fixja:competition:list")
        with self.login(self.superuser):
            self.assertGoodView("admin:fixja:competition:list")
        self.assertGoodNamespace(competition)

    def test_clubrole(self):
        role = factories.ClubRoleFactory.create()
        self.assertGoodNamespace(role, data={"name": "Club Role"})

    def test_teamrole(self):
        role = factories.TeamRoleFactory.create()
        self.assertGoodNamespace(role, data={"name": "Team Role"})

    def test_season(self):
        season = factories.SeasonFactory.create()
        self.assertGoodNamespace(season)

    def test_season_exclusion(self):
        exclusion = factories.SeasonExclusionDateFactory.create()
        self.assertGoodNamespace(exclusion)

    def test_timeslot(self):
        matchtime = factories.SeasonMatchTimeFactory.create()
        self.assertGoodNamespace(matchtime)

    def test_venue(self):
        venue = factories.VenueFactory.create()
        self.assertGoodNamespace(venue)

    def test_ground(self):
        ground = factories.GroundFactory.create()
        self.assertGoodNamespace(ground)

    def test_division(self):
        division = factories.DivisionFactory.create()
        self.assertGoodNamespace(division)

    def test_division_exclusion(self):
        exclusion = factories.DivisionExclusionDateFactory.create()
        self.assertGoodNamespace(exclusion)

    def test_team(self):
        team = factories.TeamFactory.create()
        self.assertGoodNamespace(team)

    def test_teamassociation(self):
        team_association = factories.TeamAssociationFactory.create()
        self.assertGoodNamespace(team_association)

    def test_stage(self):
        stage = factories.StageFactory.create()
        self.assertGoodNamespace(stage, test_query_count=100)

    def test_match(self):
        match = factories.MatchFactory.create()
        self.assertGoodNamespace(match)

    @unittest.skipIf(
        VERSION > (4, 0), "FIXME: Django 4.1+ error seems to be factory_boy related"
    )
    def test_pool(self):
        pool = factories.StageGroupFactory.create()
        self.assertGoodNamespace(pool)

    @unittest.skip(
        "django-guardian bug - see "
        "https://github.com/django-guardian/django-guardian/issues/519"
    )
    def test_club(self):
        club = factories.ClubFactory.create()
        self.assertLoginRequired("admin:fixja:club:list")
        with self.login(self.superuser):
            self.assertGoodView("admin:fixja:club:list")
        self.assertGoodNamespace(club)

    def test_person(self):
        person = factories.PersonFactory.create()
        self.assertFalse(person.statistics.count())
        self.assertGoodNamespace(person)

    def test_format(self):
        draw_format = factories.DrawFormatFactory.create()
        self.assertLoginRequired("admin:fixja:format:list")
        with self.login(self.superuser):
            self.assertGoodView("admin:fixja:format:list")
        self.assertGoodNamespace(draw_format)

    @unittest.skip("Not refactored as a namespace yet")
    def test_list_drawformat(self):
        self.assertLoginRequired("admin:fixja:report:list")
        with self.login(self.superuser):
            self.assertGoodView("admin:fixja:report:list")

    # code coverage

    def test_season_report(self):
        season = factories.SeasonFactory.create()
        self.assertLoginRequired(
            "admin:fixja:competition:season:report", season.competition_id, season.pk
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:competition:season:report",
                season.competition_id,
                season.pk,
            )

    def test_season_summary(self):
        season = factories.SeasonFactory.create()
        self.assertLoginRequired(
            "admin:fixja:competition:season:summary", season.competition_id, season.pk
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:competition:season:summary",
                season.competition_id,
                season.pk,
            )

    def test_match_schedule_season(self):
        season = factories.SeasonFactory.create()
        self.assertLoginRequired(
            "admin:fixja:match-schedule", season.competition_id, season.pk, "20170213"
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:match-schedule",
                season.competition_id,
                season.pk,
                "20170213",
            )

    def test_match_schedule_division(self):
        division = factories.DivisionFactory.create()
        self.assertLoginRequired(
            "admin:fixja:match-schedule",
            competition_id=division.season.competition_id,
            season_id=division.season_id,
            datestr="20170213",
            division_id=division.pk,
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:match-schedule",
                competition_id=division.season.competition_id,
                season_id=division.season_id,
                datestr="20170213",
                division_id=division.pk,
            )

    def test_progress_teams(self):
        division = factories.DivisionFactory.create()
        teams = factories.TeamFactory.create_batch(division=division, size=6)

        round_games, finals = factories.StageFactory.create_batch(
            division=division, size=2
        )

        for round in round_robin(teams):
            for home_team, away_team in round:
                factories.MatchFactory.create(
                    stage=round_games,
                    home_team=home_team,
                    away_team=away_team,
                )

        # Easier to just build the 3 matches required for progression than to
        # automate it more extensively.
        semi_1 = factories.MatchFactory.create(
            stage=finals,
            label="Semi 1",
            round=6,
            home_team=None,
            home_team_eval="P1",
            away_team=None,
            away_team_eval="P4",
        )
        semi_2 = factories.MatchFactory.create(
            stage=finals,
            label="Semi 2",
            round=6,
            home_team=None,
            home_team_eval="P2",
            away_team=None,
            away_team_eval="P3",
        )
        factories.MatchFactory.create(
            stage=finals,
            label="Final",
            round=7,
            home_team=None,
            home_team_eval="W",
            home_team_eval_related=semi_1,
            away_team=None,
            away_team_eval="W",
            away_team_eval_related=semi_2,
        )

        self.assertLoginRequired(
            round_games._get_admin_namespace() + ":draw:progress",
            *round_games._get_url_args(),
        )

        with self.login(self.superuser):
            # The round games cannot be progressed, it's the first stage and
            # therefore can't have any undecided teams.
            self.get(
                round_games._get_admin_namespace() + ":draw:progress",
                *round_games._get_url_args(),
            )
            self.response_410()

            # FIXME
            #
            # # The final series can't be progressed because there are results
            # # that need to be entered for matches in the preceding stage.
            # self.get(finals._get_admin_namespace() + ':draw:progress',
            #          *finals._get_url_args())
            # self.response_410()
            #
            # # Set random results for every unplayed match in the previous stage
            # # so that we can determine progressions.
            # for m in round_games.matches.all():
            #     m.home_team_score = random.randint(0, 20)
            #     m.away_team_score = random.randint(0, 20)

            # The final series progression should now be accessible.
            self.get(
                finals._get_admin_namespace() + ":draw:progress",
                *finals._get_url_args(),
            )
            self.response_200()

    def test_edit_person(self):
        person = factories.PersonFactory.create()
        self.assertLoginRequired("admin:fixja:club:person:add", person.club.pk)
        self.assertLoginRequired(
            "admin:fixja:club:person:edit", person.club.pk, person.pk
        )
        with self.login(self.superuser):
            self.assertGoodView("admin:fixja:club:person:add", person.club.pk)
            self.assertGoodView(
                "admin:fixja:club:person:edit", person.club.pk, person.pk
            )

    def test_merge_person(self):
        "Create two people, destructively merge second into the first."
        club = factories.ClubFactory.create()

        # Not using the PersonFactory because it requires a User, we want a
        # bare Person instance without that relationship.
        member1 = club.members.create(first_name="Alice", last_name="User")

        # Our second Person should have the User property, because we want to
        # transfer that to the first.
        member2 = factories.PersonFactory.create(club=club)

        # Check our initial setup is as we expect.
        self.assertCountEqual(
            [
                ("Alice", "User", None),
                (member2.first_name, member2.last_name, member2.user.pk),
            ],
            club.members.values_list("first_name", "last_name", "user"),
        )

        # Exercise the merge view.
        with self.login(self.superuser):
            data = {
                "first_name": "Alice",
                "last_name": "User",
                "gender": "F",
                "date_of_birth": "1983-04-27",
                "email": "",
                "home_phone": "",
                "work_phone": "",
                "mobile_phone": "",
                "user": "",
                "merge": member2.pk,
                "keep_old": "false",
            }
            self.post("admin:fixja:club:person:merge", club.pk, member1.pk, data=data)
            self.response_302()

        # Check the database state after the merge.
        self.assertCountEqual(
            [
                ("Alice", "User", member2.user.pk, "F", date(1983, 4, 27)),
            ],
            club.members.values_list(
                "first_name", "last_name", "user", "gender", "date_of_birth"
            ),
        )

    def test_edit_clubassociation(self):
        association = factories.ClubAssociationFactory.create()
        self.assertLoginRequired(
            "admin:fixja:club:clubassociation:add", association.club.pk
        )
        self.assertLoginRequired(
            "admin:fixja:club:clubassociation:edit", association.club.pk, association.pk
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:club:clubassociation:add", association.club.pk
            )
            self.assertGoodView(
                "admin:fixja:club:clubassociation:edit",
                association.club.pk,
                association.pk,
            )

    # def test_delete_competition(self):
    #     self.assertLoginRequired('admin:fixja:competition:delete', 1)
    #     with self.login(username='gary@touch.asn.au'):
    #         self.get('admin:fixja:competition:delete', 1)
    #         self.response_405()
    #         self.post('admin:fixja:competition:delete', 1)
    #         self.response_302()

    def test_perms_views(self):
        team = factories.TeamFactory.create()
        self.assertLoginRequired(
            team.division.season.competition._get_admin_namespace() + ":perms",
            *team._get_url_args()[:1],
        )
        self.assertLoginRequired(
            team.division.season._get_admin_namespace() + ":perms",
            *team._get_url_args()[:2],
        )
        # self.assertLoginRequired(
        #     team.division._get_admin_namespace() + ':perms',
        #     *team._get_url_args()[:3])
        self.assertLoginRequired(
            team._get_admin_namespace() + ":perms", *team._get_url_args()
        )

        with self.login(self.superuser):
            self.assertGoodView(
                team.division.season.competition._get_admin_namespace() + ":perms",
                *team._get_url_args()[:1],
            )
            self.assertGoodView(
                team.division.season._get_admin_namespace() + ":perms",
                *team._get_url_args()[:2],
            )
            # self.assertGoodView(
            #     team.division._get_admin_namespace() + ':perms',
            #     *team._get_url_args()[:3])
            with self.assertNumQueriesLessThan(100):
                self.get(team._get_admin_namespace() + ":perms", *team._get_url_args())
            self.response_200()

    def test_registration_form(self):
        club = factories.ClubFactory.create()
        team = factories.TeamFactory.create(club=club)
        team.division.season.competition.clubs.add(club)
        self.assertLoginRequired(
            "admin:fixja:club:registration-form",
            club.pk,
            team.division.season.pk,
            "html",
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:club:registration-form",
                club.pk,
                team.division.season.pk,
                "html",
            )

    @unittest.expectedFailure
    def test_edit_team_members(self):
        club = factories.ClubFactory.create()
        team = factories.TeamFactory.create(club=club)
        self.assertLoginRequired(
            "admin:fixja:club:team:edit", club.pk, team.division.season.pk, team.pk
        )
        with self.login(self.superuser):
            self.assertGoodView(
                "admin:fixja:club:team:edit", club.pk, team.division.season.pk, team.pk
            )


class BackendTests(TestCase):
    def setUp(self):
        super(BackendTests, self).setUp()
        self.login(self.superuser)

    def test_add_division(self):
        """
        Using the admin interface, add a new division and check that it appears
        in the front-end. Later update the form data and repost to ensure the
        front-end is also updated.
        """
        season = factories.SeasonFactory.create()
        data = {
            "title": "Mixed 4",
            "short_title": "X4",
            "points_formula_0": "3",
            "points_formula_1": "2",
            "points_formula_2": "1",
            "points_formula_3": "",
            "points_formula_4": "",
            "points_formula_5": "",
            "forfeit_for_score": "5",
            "forfeit_against_score": "0",
            "include_forfeits_in_played": "1",
            "slug": "",
            "slug_locked": "0",
        }
        self.post(
            season._get_admin_namespace() + ":division:add",
            *season._get_url_args(),
            data=data,
        )
        self.response_302()

        with self.settings(ROOT_URLCONF="tournamentcontrol.competition.tests.urls"):
            self.assertGoodView(
                "competition:season", season.competition.slug, season.slug
            )
            self.assertResponseContains("Mixed 4", html=False)
            self.assertGoodView(
                "competition:division", season.competition.slug, season.slug, "mixed-4"
            )
            self.assertResponseContains("Mixed 4", html=False)

        data.update(
            {
                "title": "4th Division Mixed",
            }
        )
        division = season.divisions.latest("pk")

        self.post(
            division._get_admin_namespace() + ":edit",
            *division._get_url_args(),
            data=data,
        )
        self.response_302()

        with self.settings(ROOT_URLCONF="tournamentcontrol.competition.tests.urls"):
            self.assertGoodView(
                "competition:season", season.competition.slug, season.slug
            )
            self.assertResponseNotContains("Mixed 4", html=False)
            self.assertResponseContains("4th Division Mixed", html=False)

    def test_update_club(self):
        """
        Using the admin interface, update an existing Club and ensure that the
        revised attribute is correctly updated.
        """
        club = factories.ClubFactory.create()
        data = {
            "title": "Bare Back Riders",
            "short_title": "",
            "abbreviation": "BBR",
            "website": "",
            "twitter": "",
            "primary": "",
            "slug": "bare-back-riders",
            "slug_locked": "0",
        }
        self.post("admin:fixja:club:edit", club.pk, data=data)
        self.response_302()

        self.get("admin:fixja:club:list")
        self.assertResponseContains("Bare Back Riders", html=False)
        self.assertResponseContains("BBR", html=False)

    def test_match_detail(self):
        play_at = factories.GroundFactory.create()
        m = factories.MatchFactory.create(
            play_at=play_at, home_team_score=3, away_team_score=0
        )

        # Make sure home team has some players associated with it.
        factories.TeamAssociationFactory.create_batch(
            4, is_player=True, team=m.home_team, person__club=m.home_team.club
        )

        kwargs = {
            "match_id": m.pk,
            "stage_id": m.stage_id,
            "division_id": m.stage.division_id,
            "season_id": m.stage.division.season_id,
            "competition_id": m.stage.division.season.competition_id,
        }
        url = reverse(
            "admin:competition:competition:season:division:stage:match:detail",
            kwargs=kwargs,
        )

        res = self.client.get(url)

        # After GET ensure away_team has no players
        text = """
        <td colspan="5" class="no_results">
            There are no registered players in this team.
        </td>
        """
        self.assertContains(res, text, html=True)

        data = {
            "home-0-number": "3",
            "home-0-id": "",
            "home-0-played": "1",
            "home-0-points": "0",
            "home-0-mvp": "0",
            "home-1-number": "8",
            "home-1-id": "",
            "home-1-played": "1",
            "home-1-points": "0",
            "home-1-mvp": "0",
            "home-2-number": "10",
            "home-2-id": "",
            "home-2-played": "1",
            "home-2-points": "0",
            "home-2-mvp": "0",
            "home-3-number": "12",
            "home-3-id": "",
            "home-3-played": "1",
            "home-3-points": "0",
            "home-3-mvp": "0",
            "home-TOTAL_FORMS": "4",
            "home-INITIAL_FORMS": "4",
            "home-MAX_NUM_FORMS": "1000",
            "away-TOTAL_FORMS": "0",
            "away-INITIAL_FORMS": "0",
            "away-MAX_NUM_FORMS": "1000",
        }

        res = self.client.post(url, data=data)
        text = """
        <ul class="list-group">
            <li class="list-group-item list-group-item-danger">
                Total number of points (0) does not equal total
                number of scores (3) for this team.
            </li>
        </ul>
        """
        self.assertContains(res, text, html=True)

        data.update(
            {
                "home-0-points": "0",
                "home-1-points": "1",
                "home-2-points": "0",
                "home-3-points": "2",
            }
        )
        res = self.client.post(url, data=data)
        self.assertRedirects(res, reverse("admin:index"))

        # Make sure the database is correctly reflecting the detailed entries.
        self.assertEqual(m.statistics.count(), 4)

    def test_bug_0068(self):
        """
        While there are no LadderEntry or LadderSummary records for matches
        revealed in the ``match-results`` view, the submission worked fine.

        However on subsequent processing the method decorator is invoked
        differently, and there is now "raw" keyword argument - this resulted
        in issue #68.
        """
        match = factories.MatchFactory.create()
        data = {
            # first formset
            "matches-INITIAL_FORMS": "1",
            "matches-MAX_NUM_FORMS": "1000",
            "matches-TOTAL_FORMS": "1",
            "matches-0-id": str(match.pk),
            "matches-0-home_team_score": "4",
            "matches-0-away_team_score": "3",
            "matches-0-is_forfeit": "0",
            # second formset
            "byes-INITIAL_FORMS": "0",
            "byes-MAX_NUM_FORMS": "1000",
            "byes-TOTAL_FORMS": "0",
        }

        self.post(
            "admin:fixja:match-results",
            match.stage.division.season.competition.pk,
            match.stage.division.season.pk,
            match.date.strftime("%Y%m%d"),
            data=data,
        )
        self.response_302()

        data.update(
            {
                "matches-1-home_team_score": "5",
                "matches-1-away_team_score": "2",
            }
        )
        self.post(
            "admin:fixja:match-results",
            match.stage.division.season.competition.pk,
            match.stage.division.season.pk,
            match.date.strftime("%Y%m%d"),
            data=data,
        )
        self.response_302()

    def test_enhancement_0024_team(self):
        """
        When editing a Team or UndecidedTeam it should be possible to delete
        the object if there are no matches that team has been assigned to.
        """
        stage = factories.StageFactory.create()
        home = factories.TeamFactory.create(division=stage.division)
        away = factories.UndecidedTeamFactory.create(stage=stage)

        factories.MatchFactory.create(
            stage=stage, home_team=home, away_team=None, away_team_undecided=away
        )

        delete_url = self.reverse(
            home._get_admin_namespace() + ":delete", *home._get_url_args()
        )

        # Ensure the edit view response does not include a delete button.
        self.get(home._get_admin_namespace() + ":edit", *home._get_url_args())
        self.assertResponseNotContains(delete_url, html=False)

        # Ensure that visiting the delete view does not respond when matches
        # are associated with the Team.
        self.get(home._get_admin_namespace() + ":delete", *home._get_url_args())
        self.response_410()

        # Ensure that POST to the delete view fails the same as for GET.
        self.post(home._get_admin_namespace() + ":delete", *home._get_url_args())
        self.response_410()

        # Create a new Team and re-check the scenarios.
        team = factories.TeamFactory.create(division=stage.division)

        delete_url = self.reverse(
            team._get_admin_namespace() + ":delete", *team._get_url_args()
        )

        # Ensure the edit view response does not include a delete button.
        self.get(team._get_admin_namespace() + ":edit", *team._get_url_args())
        self.assertResponseNotContains(delete_url, html=False)

        # Ensure that visiting the delete view is not allowed when there are no
        # matches but you use a GET request.
        self.get(team._get_admin_namespace() + ":delete", *team._get_url_args())
        self.response_405()

        # Ensure that POST to the delete view redirects.
        self.post(team._get_admin_namespace() + ":delete", *team._get_url_args())
        self.response_302()

        # Subsequent GET request to the edit view should be a 404.
        self.get(team._get_admin_namespace() + ":edit", *team._get_url_args())
        self.response_404()

    def test_enhancement_0024_undecidedteam(self):
        """
        When editing a Team or UndecidedTeam it should be possible to delete
        the object if there are no matches that team has been assigned to.
        """
        stage = factories.StageFactory.create()
        home = factories.TeamFactory.create(division=stage.division)
        away = factories.UndecidedTeamFactory.create(stage=stage)

        factories.MatchFactory.create(
            stage=stage, home_team=home, away_team=None, away_team_undecided=away
        )

        delete_url = self.reverse(
            away._get_admin_namespace() + ":delete", *away._get_url_args()
        )

        # Ensure the edit view response does not include a delete button.
        self.get(away._get_admin_namespace() + ":edit", *away._get_url_args())
        self.assertResponseNotContains(delete_url, html=False)

        # Ensure that visiting the delete view does not respond when matches
        # are associated with the Team.
        self.get(away._get_admin_namespace() + ":delete", *away._get_url_args())
        self.response_410()

        # Ensure that POST to the delete view fails the same as for GET.
        self.post(away._get_admin_namespace() + ":delete", *away._get_url_args())
        self.response_410()

        # Create a new UndecidedTeam and re-check the scenarios.
        team = factories.UndecidedTeamFactory.create(stage=stage)

        delete_url = self.reverse(
            team._get_admin_namespace() + ":delete", *team._get_url_args()
        )

        # Ensure the edit view response *does* include a delete button.
        self.get(stage._get_admin_namespace() + ":edit", *stage._get_url_args())
        self.assertResponseContains(delete_url, html=False)

        # Ensure that visiting the delete view is not allowed when there are no
        # matches but you use a GET request.
        self.get(team._get_admin_namespace() + ":delete", *team._get_url_args())
        self.response_405()

        # Ensure that POST to the delete view redirects.
        self.post(team._get_admin_namespace() + ":delete", *team._get_url_args())
        self.response_302()

        # Subsequent GET request to the edit view should be a 404.
        self.get(team._get_admin_namespace() + ":delete", *team._get_url_args())
        self.response_404()

    @unittest.skipIf(
        VERSION > (4, 0), "FIXME: Django 4.1+ error seems to be factory_boy related"
    )
    def test_bug_80_add_stagegroup(self):
        "Add Pool should gain order equal to max(order) + 1"
        pool = factories.StageGroupFactory.create()
        data = {
            "title": "Group Z",
            "short_title": "",
            "slug": "",
            "slug_locked": "0",
            "carry_ladder": "0",
        }
        self.post(
            pool._get_admin_namespace() + ":add", *pool._get_url_args()[:-1], data=data
        )
        self.response_302()
        z = pool.stage.pools.latest("order")
        self.assertEqual(z.order, pool.order + 1)

    def test_bug_0115_add_duplicate_team_name(self):
        """
        When adding two teams to the same division, it should not allow the
        second one because of database integrity constraints.
        """
        team = factories.TeamFactory.create()
        data = {
            "title": team.title,
        }
        self.post(
            team._get_admin_namespace() + ":add", *team._get_url_args()[:-1], data=data
        )
        form = self.get_context("form")
        self.assertFormError(form, "title", ["Team with this Title already exists."])

    def test_team_add(self):
        division = factories.DivisionFactory.create()
        data = {
            "title": "New Team",  # current value
            "short_title": "",
            "names_locked": "0",
            "timeslots_after": "",
            "timeslots_before": "",
            "team_clashes": [],
        }
        self.post(
            division._get_admin_namespace() + ":team:add",
            *division._get_url_args(),
            data=data,
        )
        self.response_302()
        # Ensure that the team was created and that it has a slug
        team = Team.objects.get(title="New Team")
        self.assertEqual(team.slug, "new-team")

    def test_team_edit(self):
        team = factories.TeamFactory.create()
        data = {
            "title": "Blue",
            "short_title": "",
            "names_locked": "0",
            "timeslots_after": "19:20",  # new value
            "timeslots_before": "",
            "team_clashes": [],
        }

        self.post(
            team._get_admin_namespace() + ":edit", *team._get_url_args(), data=data
        )
        self.response_302()

        # Ensure that the team was updated
        blue = Team.objects.get(title="Blue")
        self.assertEqual(team.pk, blue.pk)
        self.assertEqual(blue.slug, "blue")
        self.assertEqual(blue.timeslots_after, time(19, 20))

        # Try again, update the title and make sure the slug changes
        data["title"] = "Magenta"
        self.post(
            blue._get_admin_namespace() + ":edit", *blue._get_url_args(), data=data
        )
        self.response_302()

        magenta = Team.objects.get(title="Magenta")
        self.assertEqual(team.pk, magenta.pk)
        self.assertEqual(magenta.slug, "magenta")
        self.assertEqual(magenta.timeslots_after, time(19, 20))

    def test_add_stage_duplicate_slug(self):
        """
        When adding a new stage to a division it should not be possible to add
        a second stage with a slug already used because of database integrity
        constraints.
        """
        stage = factories.StageFactory.create()
        data = {
            "title": stage.title,
        }
        self.post(
            stage._get_admin_namespace() + ":add",
            *stage._get_url_args()[:-1],
            data=data,
        )
        form = self.get_context("form")
        self.assertFormError(form, "title", ["Stage with this Title already exists."])

    def test_bug_85_add_match_season_live_stream(self):
        """
        When the Season is set to live stream, the Match form changes.

        The ability to manually add a match under this configuration should be possible,
        but there is a bug caused by accessing unset relations.

        Refs: https://github.com/goodtune/vitriolic/issues/85
        """
        stage = factories.StageFactory.create(division__season__live_stream=True)
        team1 = factories.TeamFactory.create(division=stage.division)
        team2 = factories.TeamFactory.create(division=stage.division)
        data = {
            "home_team": team1.pk,
            "away_team": team2.pk,
            "label": "",
            "round": 1,
            "date": "2025-04-28",
            "include_in_ladder": "0",
            "live_stream": "0",
            "thumbnail_url": "",
        }
        self.post(
            stage._get_admin_namespace() + ":match:add",
            *stage._get_url_args(),
            data=data,
        )
        self.response_302()

    def test_bug_100_add_stagegroup(self):
        """
        Adding a Pool (StageGroup) should not fail to render.
        """
        stage = factories.StageFactory.create()
        self.get_check_200(
            stage._get_admin_namespace() + ":stagegroup:add",
            *stage._get_url_args(),
        )



import logging

from django.contrib.auth.decorators import (
    login_required,
    permission_required,
    user_passes_test,
)
from django.utils.decorators import method_decorator
from django.utils.functional import wraps
from django.views.decorators.cache import cache_control, never_cache
from django.views.decorators.csrf import csrf_exempt
from django.views.decorators.http import require_GET, require_POST

logger = logging.getLogger(__name__)


@user_passes_test
def staff_login_required(user):
    logger.debug("staff_login_required: %s: %s", user, user.is_staff)
    return user.is_staff


@user_passes_test
def superuser_login_required(user):
    logger.debug("superuser_login_required: %s: %s", user, user.is_superuser)
    return user.is_superuser


def node2extracontext(f, *a, **kw):
    @wraps(f)
    def _decorated(*args, **kwargs):
        node = kwargs.pop("node", None)
        extra_context = kwargs.setdefault("extra_context", {})
        if node is not None:
            extra_context.update({"node": node})
        return f(*args, **kwargs)

    return _decorated


cache_control_m = method_decorator(cache_control)
csrf_exempt_m = method_decorator(csrf_exempt)
login_required_m = method_decorator(login_required)
never_cache_m = method_decorator(never_cache)
permission_required_m = method_decorator(permission_required)
staff_login_required_m = method_decorator(staff_login_required)
superuser_login_required_m = method_decorator(superuser_login_required)
user_passes_test_m = method_decorator(user_passes_test)

require_GET_m = method_decorator(require_GET)
require_POST_m = method_decorator(require_POST)



"""
Django settings for vitriolic project.
"""

import os
import time

import environ
from django.urls import reverse_lazy

env = environ.Env()

# Build paths inside the project like this: os.path.join(BASE_DIR, ...)
BASE_DIR = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"
SILENCED_SYSTEM_CHECKS = env.list("SILENCED_SYSTEM_CHECKS", default=[])
SILENCED_SYSTEM_CHECKS.append("models.E006")

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/1.10/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = "3h_k7=3wv&i&#^36t=zv)l99bijpp06j((ld%7u7&3u)2!8iq8"

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = True

ALLOWED_HOSTS = []

SITE_ID = 1


class DisableMigrations(object):
    """
    Disable the database migrations machinery to speed up test suite.

    Trick learned at http://bit.ly/2vjVpNc
    """

    def __contains__(self, item):
        return True

    def __getitem__(self, item):
        return None


MIGRATION_MODULES = DisableMigrations()


# Application definition

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.humanize",
    "django.contrib.messages",
    "django.contrib.postgres",
    "django.contrib.sessions",
    "django.contrib.sites",
    "django.contrib.staticfiles",
    "mptt",
    "cloudinary",
    "guardian",
    "bootstrap3",
    "django_gravatar",
    "embed_video",
    "touchtechnology.common",
    "touchtechnology.admin",
    "touchtechnology.content",
    "touchtechnology.news",
    "tournamentcontrol.competition",
    "example_app",
]

MIDDLEWARE = [
    "django.contrib.sites.middleware.CurrentSiteMiddleware",
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
    "touchtechnology.common.middleware.served_by_middleware",
    "touchtechnology.content.middleware.SitemapNodeMiddleware",
    "touchtechnology.content.middleware.redirect_middleware",
]

ROOT_URLCONF = "vitriolic.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                # It would be better if we could use modify_settings to append
                # these as required, but all the core Django checks look like
                # overriding TEMPLATES is done as override_settings and done
                # in entirety.
                "touchtechnology.common.context_processors.env",
                "touchtechnology.common.context_processors.query_string",
                "touchtechnology.common.context_processors.site",
                "touchtechnology.common.context_processors.tz",
            ],
            "loaders": [
                "django.template.loaders.filesystem.Loader",
                "django.template.loaders.app_directories.Loader",
            ],
        },
    },
]

WSGI_APPLICATION = "vitriolic.wsgi.application"


# Database
# https://docs.djangoproject.com/en/1.10/ref/settings/#databases

DATABASES = {
    "default": env.db(default="psql://vitriolic:vitriolic@localhost/vitriolic"),
}

if DATABASES["default"]["ENGINE"].startswith("django.db.backends.postgresql"):
    DATABASES["default"]["PORT"] = env.int("DB_5432_TCP_PORT", default=5432)
    # delay long enough to let the postgresql container startup
    time.sleep(4)


# Password validation
# https://docs.djangoproject.com/en/1.10/ref/settings/#auth-password-validators

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator"
    },
    {"NAME": "django.contrib.auth.password_validation.MinimumLengthValidator"},
    {"NAME": "django.contrib.auth.password_validation.CommonPasswordValidator"},
    {"NAME": "django.contrib.auth.password_validation.NumericPasswordValidator"},
]

PASSWORD_HASHERS = [
    "django.contrib.auth.hashers.MD5PasswordHasher",
]

AUTHENTICATION_BACKENDS = (
    "django.contrib.auth.backends.ModelBackend",  # this is default
    "guardian.backends.ObjectPermissionBackend",
)

LOGIN_URL = reverse_lazy("accounts:login")

ANONYMOUS_USER_NAME = "anonymous"


# Internationalization
# https://docs.djangoproject.com/en/1.10/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_L10N = True

USE_TZ = True


# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/1.10/howto/static-files/

STATIC_URL = "/static/"


# OAuth2

GOOGLE_OAUTH2_CLIENT_ID = ""
GOOGLE_OAUTH2_CLIENT_SECRET = ""


# Logging setup. Adjust handlers as required.

LOGGING = {
    "version": 1,
    "handlers": {
        "console": {"class": "logging.StreamHandler"},
        "null": {"class": "logging.NullHandler"},
    },
    "loggers": {
        "": {"level": "DEBUG", "handlers": ["null"]},
    },
}


# Touch Technology settings

TOUCHTECHNOLOGY_SITEMAP_ROOT = "home"



# -*- coding: utf-8 -*-

from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("news", "0003_category_last_modified"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="category",
            options={"verbose_name_plural": "categories"},
        ),
        migrations.AlterField(
            model_name="article",
            name="is_active",
            field=touchtechnology.common.db.models.BooleanField(
                default=True, verbose_name="Enabled"
            ),
        ),
        migrations.AlterField(
            model_name="article",
            name="last_modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="article",
            name="published",
            field=models.DateTimeField(
                help_text="Set a date & time in the future to schedule an announcement.",
                verbose_name="Published",
            ),
        ),
        migrations.AlterField(
            model_name="article",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Locked Slug"
            ),
        ),
        migrations.AlterField(
            model_name="category",
            name="hidden_from_navigation",
            field=touchtechnology.common.db.models.BooleanField(
                default=False,
                help_text="When set to 'Yes' this object will still be available, but will not appear in menus.",
                verbose_name="Hide from menus",
            ),
        ),
        migrations.AlterField(
            model_name="category",
            name="is_active",
            field=touchtechnology.common.db.models.BooleanField(
                default=True, verbose_name="Enabled"
            ),
        ),
        migrations.AlterField(
            model_name="category",
            name="last_modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="category",
            name="slug_locked",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Locked Slug"
            ),
        ),
    ]



import logging
import os.path
from urllib.parse import urljoin

from django.conf import settings
from django.contrib import messages
from django.contrib.auth import forms, get_user_model, views
from django.contrib.auth.models import Permission
from django.contrib.contenttypes.models import ContentType
from django.contrib.postgres.search import SearchVector
from django.core.paginator import EmptyPage, Paginator
from django.db import transaction
from django.forms import ModelForm as BaseModelForm
from django.forms.models import inlineformset_factory, modelformset_factory
from django.http import HttpResponse
from django.shortcuts import get_object_or_404, redirect
from django.template import RequestContext
from django.template.loader import render_to_string
from django.template.response import TemplateResponse
from django.urls import NoReverseMatch, path, reverse, reverse_lazy
from django.utils import timezone
from django.utils.encoding import smart_str
from django.utils.http import urlencode
from django.utils.module_loading import import_string
from django.utils.translation import gettext_lazy as _
from django.views.decorators.cache import patch_cache_control
from guardian.shortcuts import get_objects_for_user
from guardian.utils import get_40x_or_None
from modelforms.forms import ModelForm

from touchtechnology.common.decorators import (
    login_required_m,
    node2extracontext,
    require_POST_m,
)
from touchtechnology.common.default_settings import (
    PAGINATE_BY,
    PROFILE_FORM_CLASS,
)
from touchtechnology.common.utils import (
    get_all_perms_for_model_cached,
    get_perms_for_model,
    model_and_manager,
    select_template_name,
)

logger = logging.getLogger(__name__)


class AlreadyRegistered(Exception):
    pass


class Application(object):
    kwargs_form_class = None
    model_form_bases = (ModelForm,)

    def __init__(self, name, app_name, node=None, **kwargs):
        self.name = name
        self.app_name = app_name
        self.node = node
        # store any instantiation keyword arguments to change behaviour
        self.kwargs = kwargs
        self._spawned = timezone.now()

    def __repr__(self):
        return "<{0}: {1}/{2}/{3}?{4}>".format(
            self.__class__.__name__,
            self.name,
            self.app_name,
            self.node or "",
            urlencode(self.kwargs),
        )

    def _get_namespace(self, cls=None):
        if cls is None:
            cls = self.__class__
        ns, __ = cls.__module__.split(".", 1)
        return ns if ns not in ("apps", self.app_name) else ""

    namespace = property(_get_namespace)

    @property
    def template_base(self):
        # configure the applications base template directory based on
        # namespace and app name
        return os.path.join(self.namespace, self.app_name)

    @property
    def urls(self):
        return self.get_urls(), self.app_name, self.name

    @classmethod
    def verbose_name(cls):
        return cls().app_name.title()

    @property
    def version(self):
        """
        Identify version number of the Application. If it is defined as a
        module we look up the import path until we find the module level
        ``get_version`` function, to a maximum depth of 4 levels.
        """
        ver = ""
        for depth in range(1, 5):
            parts = self.__module__.rsplit(".", depth)
            mod = parts[0]
            try:
                ver = import_string("%s.__version__" % mod)
            except ImportError:
                logger.warning('"__version__" not found at "%s"', mod)
            else:
                logger.debug('Identified version "%s" for "%s"', ver, mod)
                break
        name = self.verbose_name
        if callable(name):
            name = name()
        return "{name} {ver}".format(name=_(name), ver=ver)

    @property
    def current_app(self):
        return self.name

    def reverse(self, name, args=None, kwargs=None, prefix=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        named_url = "%s:%s" % (self.app_name, name)
        if prefix:
            named_url = "%s:%s" % (prefix, named_url)
        for key in self.kwargs:
            kwargs.pop(key, None)
        return reverse_lazy(
            named_url, args=args, kwargs=kwargs, current_app=self.current_app
        )

    def redirect(self, *args, **kwargs):
        "Thin wrapper around django.shortcuts.redirect to save on imports"
        return redirect(*args, **kwargs)

    def _template_path(self, filename, *args):
        args = list(reversed([a for a in args if a]))
        for index, arg in enumerate(args, 1):
            yield os.path.join(
                self.template_base, os.path.join(*args[:index]), filename
            )
        yield os.path.join(self.template_base, filename)

    def template_path(self, *args, **kwargs):
        return list(self._template_path(*args, **kwargs))

    def generic_permissions(
        self,
        request,
        model,
        pk=None,
        instance=None,
        post_save_redirect=None,
        templates=None,
        staff_only=None,
        max_checkboxes=None,
        **extra_context,
    ):
        # avoid circular import
        from touchtechnology.common.forms.auth import permissionformset_factory

        if instance is None:
            assert pk is not None
            instance = get_object_or_404(model, pk=pk)

        perms = get_perms_for_model(model, add=True)
        has_permission = get_40x_or_None(
            request,
            perms,
            instance,
            return_403=True,
            accept_global_perms=True,
        )

        if has_permission and not request.user.is_superuser:
            return has_permission

        extra_context.update(
            {
                "model": model,
            }
        )

        content_type = ContentType.objects.get_for_model(model)
        queryset = Permission.objects.filter(content_type=content_type).exclude(
            codename__startswith="add_"
        )
        formset_class = permissionformset_factory(
            model, staff_only=staff_only, max_checkboxes=max_checkboxes
        )

        if templates is None:
            templates = self.template_path("permissions.html", model._meta.model_name)

        return self.generic_edit_multiple(
            request,
            model,
            queryset=queryset,
            formset_class=formset_class,
            formset_kwargs={
                "instance": instance,
                "queryset": queryset,
            },
            post_save_redirect=post_save_redirect,
            templates=templates,
            extra_context=extra_context,
        )

    def generic_list(
        self,
        request,
        model_or_manager,
        extra_context=None,
        object_list_name=None,
        paginate_by=None,
        queryset=None,
        templates=None,
        list_templates=None,
        perms=None,
        permission_required=False,
        accept_global_perms=True,
        return_403=None,
        search=None,
        *args,
        **kwargs,
    ):

        model, manager = model_and_manager(model_or_manager)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return an HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        if queryset is None:
            queryset = manager.all()

        # Search should be an iterable of fields to look into.
        if search is not None:
            terms = request.GET.get("search")
            if terms is not None:
                queryset = queryset.annotate(
                    search=SearchVector(*search),
                ).filter(search=terms)

        # Implement permission checking for list of objects, will filter the
        # list to those objects the user is entitled to work on if we do not
        # throw an HTTP 403 for no permission.
        if permission_required:
            if perms is None:
                perms = get_all_perms_for_model_cached(model)

            # Determine the user's permission to see the list using the
            # get_40x_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_40x_or_None(
                request,
                perms,
                return_403=return_403,
                accept_global_perms=accept_global_perms,
                any_perm=True,
            )

            global_perms = any([request.user.has_perm(p) for p in perms])

            # If the user does not have any global permissions then adjust the
            # queryset to return only the objects they can act on.
            if not global_perms:
                queryset = get_objects_for_user(
                    request.user, perms, queryset, use_groups=True, any_perm=True
                )

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if not queryset and has_permission is not None:
                return has_permission

        if extra_context is None:
            extra_context = {}

        if object_list_name is None:
            object_list_name = "{0}_list".format(model._meta.model_name)

        if templates is None:
            templates = self.template_path("list.html", model._meta.model_name)

        if list_templates is None:
            list_templates = self.template_path("list.inc.html", model._meta.model_name)
            # Always tack the mvp theme version on the end as a fallback.
            list_templates.append("mvp/list.inc.html")

        # Determine the namespace of this Application, so we can construct
        # view names to pass to the template engine to reverse CRUD URI's.
        create_url_name = "%s:add" % request.resolver_match.namespace
        create_url_kwargs = request.resolver_match.kwargs.copy()
        create_url_kwargs.pop("admin", None)
        create_url_kwargs.pop("component", None)

        # We can't leave this until the template is rendered because it throws
        # exceptions (stupid!) so we'll avoid lazy evaluation and pass None if
        # the create view doesn't exist for this model.
        try:
            create_url = reverse(
                create_url_name,
                args=request.resolver_match.args,
                kwargs=create_url_kwargs,
            )
        except NoReverseMatch:
            create_url = None

        context = {
            "queryset": queryset,
            object_list_name: queryset,
            "object_list": queryset,
            "is_paginated": False,
            "page": None,
            "page_num": None,
            "paginator": None,
            "model": manager.none(),
            "template": select_template_name(templates),
            "list_template": select_template_name(list_templates),
            # Is search enabled?
            "searchable": bool(search),
            "search": request.GET.get("search", ""),
            # Pass to template the name permissions, so we can re-use template
            # code to generically list and add/change/delete objects
            "add_perm": "%s.add_%s" % (model._meta.app_label, model._meta.model_name),
            "view_perm": "%s.view_%s" % (model._meta.app_label, model._meta.model_name),
            "change_perm": "%s.change_%s"
            % (model._meta.app_label, model._meta.model_name),
            "delete_perm": "%s.delete_%s"
            % (model._meta.app_label, model._meta.model_name),
            "create_url": create_url,
        }

        if paginate_by is None:
            paginate_by = PAGINATE_BY

        # Ensure that paginate_by is an integer to prevent ZeroDivisionError
        # inside the Django paginator steps.
        paginate_by = int(paginate_by)

        if paginate_by > 0:
            try:
                page_num = int(request.GET.get("page", 1))
            except TypeError:
                page_num = 1

            paginator = Paginator(queryset, paginate_by)

            try:
                page = paginator.page(page_num)
            except EmptyPage:
                page = paginator.page(paginator.num_pages)

            context.update(
                {
                    object_list_name: page.object_list,
                    "object_list": page.object_list,
                    "is_paginated": paginator.num_pages > 1,
                    "page": page,
                    "page_num": page_num,
                    "paginator": paginator,
                }
            )

        context.update(extra_context)
        return self.render(request, templates, context, *args, **kwargs)

    def generic_detail(
        self, request, model_or_manager, templates=None, extra_context=None, **kwargs
    ):
        model, manager = model_and_manager(model_or_manager)
        instance = get_object_or_404(manager, **kwargs)

        context = {
            "object": instance,
            model._meta.model_name: instance,
            "model": manager.none(),
        }
        context.update(extra_context or {})

        if templates is None:
            templates = self.template_path("detail.html", model._meta.model_name)

        return self.render(request, templates, context)

    def generic_edit(
        self,
        request,
        model_or_manager,
        pk=None,
        instance=None,
        form_class=None,
        form_kwargs=None,
        form_fields="__all__",
        form_widgets=None,
        pre_save_callback=lambda o: o,
        post_save_callback=lambda o: None,
        post_save_redirect=None,
        templates=None,
        changed_messages=((messages.SUCCESS, _("Your {model} has been " "saved.")),),
        unchanged_messages=((messages.INFO, _("Your {model} was not changed.")),),
        permission_required=False,
        perms=None,
        accept_global_perms=True,
        return_403=None,
        related=None,
        extra_context=None,
        always_save=False,
    ):

        model, manager = model_and_manager(model_or_manager)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return an HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        if instance is None and pk is not None:
            instance = get_object_or_404(manager, pk=pk)

        if form_kwargs is None:
            form_kwargs = {}

        if extra_context is None:
            extra_context = {}

        if post_save_redirect is None:
            redirect_to = urljoin(request.path, "..") if pk is None else request.path
            post_save_redirect = self.redirect(redirect_to)

        # Implement permission checking for specified object instance, throw
        # an HTTP 403 (using django-guardian configuration) when the user is
        # not entitled to edit the instance.
        if permission_required:

            # If no perms are specified, build sensible default using built in
            # permission types that come batteries included with Django.
            if perms is None:
                perms = get_perms_for_model(model, change=True)

                # When there is no pk value, we're doing a creation and should
                # have permission to create the object.
                if pk is None or instance.pk is None:
                    perms = get_perms_for_model(model, add=True)

            # Determine the user's permission to edit this object using the
            # get_40x_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_40x_or_None(
                request,
                perms,
                obj=instance,
                return_403=return_403,
                accept_global_perms=accept_global_perms,
            )

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if has_permission is not None:
                return has_permission

        # If the developer has not provided a custom form, then dynamically
        # construct a default ModelForm for them.
        if form_class is None:
            meta_class = type(
                smart_str("Meta"),
                (),
                {"model": model, "fields": form_fields, "widgets": form_widgets},
            )
            form_class = type(
                smart_str("EditForm"), self.model_form_bases, {"Meta": meta_class}
            )

        # Whether we've dynamically constructed our form_class or not, check to
        # ensure that we've inherited from all the bases. Log when we haven't,
        # but don't raise any exceptions.
        for base in self.model_form_bases:
            if not issubclass(form_class, base):
                logger.error('"%s" does not inherit "%s"', form_class, base)

        # Pass the instance to the form constructor if this is a ModelForm
        # subclass, otherwise it will need to be explicitly added to
        # form_kwargs if expected.
        if issubclass(form_class, BaseModelForm) and instance is not None:
            form_kwargs.setdefault("instance", instance)

        # Vanilla form processing here, take the post data and files, create
        # an instance of form_class, save and redirect.
        if request.method == "POST":
            form = form_class(data=request.POST, files=request.FILES, **form_kwargs)
            if form.is_valid():
                replace = dict(
                    model=smart_str(model._meta.verbose_name),
                    models=smart_str(model._meta.verbose_name_plural),
                )
                if always_save or form.has_changed():
                    res = pre_save_callback(form.instance)
                    if isinstance(res, HttpResponse):
                        return res
                    obj = form.save()
                    callback_res = post_save_callback(obj)
                    for level, message in changed_messages:
                        messages.add_message(request, level, message.format(**replace))
                    if callback_res is not None:
                        return callback_res
                else:
                    for level, message in unchanged_messages:
                        messages.add_message(request, level, message.format(**replace))
                return post_save_redirect
        else:
            form = form_class(**form_kwargs)

        if templates is None:
            templates = self.template_path("edit.html", model._meta.model_name)
            templates.append("edit.html")

        context = {
            "model": manager.none(),
            "template": select_template_name(templates),
            "form": form,
            "object": instance,
            "related": related,
            "cancel_url": post_save_redirect.url,
        }
        context.update(extra_context or {})

        return self.render(request, templates, context)

    def generic_edit_multiple(
        self,
        request,
        model_or_manager,
        extra_context=None,
        formset_class=None,
        formset_kwargs=None,
        formset_fields="__all__",
        post_save_callback=lambda o: None,
        post_save_redirect_args=None,
        post_save_redirect_kwargs=None,
        post_save_redirect=None,
        templates=None,
        changed_messages=None,
        perms=None,
        permission_required=False,
        accept_global_perms=True,
        return_403=None,
        *args,
        **kwargs,
    ):

        if extra_context is None:
            extra_context = {}

        if formset_kwargs is None:
            formset_kwargs = {}

        if post_save_redirect_args is None:
            post_save_redirect_args = ()

        if post_save_redirect_kwargs is None:
            post_save_redirect_kwargs = {}

        if changed_messages is None:
            changed_messages = (
                (messages.SUCCESS, _("Your {models} have been saved.")),
            )

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return an HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        # Try and introspect the model if possible.
        model, manager = model_and_manager(model_or_manager)
        queryset = manager.all()

        # Implement permission checking for list of objects, will filter the
        # list to those objects the user is entitled to work on if we do not
        # throw an HTTP 403 for no permission.
        if permission_required:
            if perms is None:
                perms = get_all_perms_for_model_cached(model)

            # Determine the user's permission to see the list using the
            # get_40x_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_40x_or_None(
                request,
                perms,
                return_403=return_403,
                accept_global_perms=accept_global_perms,
                any_perm=True,
            )

            # If the user does not have any global permissions then adjust the
            # queryset to return only the objects they can act on.
            if not any([request.user.has_perm(p) for p in perms]):
                queryset = get_objects_for_user(
                    request.user, perms, queryset, use_groups=True, any_perm=True
                )

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if not queryset and has_permission is not None:
                return has_permission

            queryset = get_objects_for_user(
                request.user, perms, queryset, any_perm=True
            )

        # If queryset is not already specified as a keyword argument to the
        # formset attach it now. Value in formset_kwargs will take precedence.
        queryset = formset_kwargs.pop("queryset", queryset)

        # If the developer has not provided a custom formset, then dynamically
        # construct a default ModelFormSet for them.
        if formset_class is None:
            formset_class = modelformset_factory(model, extra=0, fields=formset_fields)

        if post_save_redirect is None:
            redirect_to = urljoin(request.path, "..")
            post_save_redirect = self.redirect(redirect_to)

        # Vanilla formset processing here, take the post data and files, create
        # an instance of formset_class, save and redirect.
        if request.method == "POST":
            formset = formset_class(
                data=request.POST,
                files=request.FILES,
                queryset=queryset,
                **formset_kwargs,
            )
            if formset.is_valid():
                replace = dict(
                    model=smart_str(model._meta.verbose_name),
                    models=smart_str(model._meta.verbose_name_plural),
                )
                objects = formset.save()
                for each in objects:
                    post_save_callback(each)
                for level, message in changed_messages:
                    messages.add_message(request, level, message.format(**replace))
                if isinstance(post_save_redirect, HttpResponse):
                    return post_save_redirect
                redirect_to = self.reverse(
                    post_save_redirect,
                    args=post_save_redirect_args,
                    kwargs=post_save_redirect_kwargs,
                )
                return redirect(redirect_to)
        else:
            formset = formset_class(queryset=queryset, **formset_kwargs)

        context = {
            "model": manager.none(),
            "formset": formset,
            "object_list": queryset,
        }
        context.update(extra_context)

        if templates is None:
            templates = self.template_path("edit_multiple.html", model._meta.model_name)
            templates.append("edit_multiple.html")

        return self.render(request, templates, context)

    def generic_edit_related(
        self,
        request,
        model_or_manager,
        related_model,
        extra_context={},
        form_class=None,
        form_kwargs={},
        form_fields="__all__",
        form_widgets=None,
        formset_class=None,
        formset_kwargs={},
        formset_fields="__all__",
        instance=None,
        pk=None,
        post_save_callback=lambda o: None,
        post_save_redirect_args=(),
        post_save_redirect_kwargs={},
        post_save_redirect="index",
        post_save_related_callback=lambda o: None,
        templates=None,
    ):

        model, manager = model_and_manager(model_or_manager)

        if instance is None and pk is not None:
            instance = get_object_or_404(manager, pk=pk)

        # If the developer has not provided a custom form, then dynamically
        # construct a default ModelForm for them.
        if form_class is None:
            meta_class = type(
                smart_str("Meta"),
                (),
                {"model": model, "fields": form_fields, "widgets": form_widgets},
            )
            form_class = type(
                smart_str("EditForm"), self.model_form_bases, {"Meta": meta_class}
            )

        # If the developer has not provided a custom formset, then dynamically
        # construct a default one with inlineformset_factory.
        if formset_class is None:
            formset_class = inlineformset_factory(
                model, related_model, fields=formset_fields, extra=0
            )

        # Add form and formset prefixes if not already present in the kwargs
        form_kwargs.setdefault("prefix", "form")
        formset_kwargs.setdefault("prefix", "formset")

        # Keep verbose names for model and related_model for use in messages
        model_verbose_name = model._meta.verbose_name.lower()
        related_verbose_name_plural = related_model._meta.verbose_name_plural.lower()
        verbose = {
            "model_verbose_name": model_verbose_name,
            "related_verbose_name_plural": related_verbose_name_plural,
        }

        # Vanilla form processing here, take the post data and files, create
        # an instance of form_class, save and redirect.
        if request.method == "POST":
            sid = transaction.savepoint()

            form = form_class(
                data=request.POST, files=request.FILES, instance=instance, **form_kwargs
            )

            if form.is_valid():
                obj = form.save()
                logger.debug("Saved %r to database.", obj)

                # Rebuild formset with updated instance object.
                formset = formset_class(
                    data=request.POST,
                    files=request.FILES,
                    instance=obj,
                    **formset_kwargs,
                )

                if formset.is_valid():
                    logger.debug(
                        "Related formset is valid, if we don't "
                        "complete successfully then an exception "
                        "will bubble up."
                    )
                    related = formset.save()
                    post_save_callback(obj)
                    for each in related:
                        post_save_related_callback(each)
                    transaction.savepoint_commit(sid)
                    messages.add_message(
                        request,
                        messages.SUCCESS,
                        _(
                            "Your %(model_verbose_name)s and "
                            "%(related_verbose_name_plural)s "
                            "have been saved."
                        )
                        % verbose,
                    )
                    if isinstance(post_save_redirect, HttpResponse):
                        return post_save_redirect
                    redirect_to = self.reverse(
                        post_save_redirect,
                        args=post_save_redirect_args,
                        kwargs=post_save_redirect_kwargs,
                    )
                    return redirect(redirect_to)

                # if we've made it out without returning a HttpRedirect then
                # we should not be persisting the result.
                logger.debug("ROLLBACK, we were not redirected.")
                transaction.savepoint_rollback(sid)

            else:
                formset = formset_class(
                    data=request.POST,
                    files=request.FILES,
                    instance=instance,
                    **formset_kwargs,
                )

        else:
            form = form_class(instance=instance, **form_kwargs)
            formset = formset_class(instance=instance, **formset_kwargs)

        context = {
            "model": manager.none(),
            "form": form,
            "formset": formset,
            "object": instance,
            "media": form.media + formset.media,
            # For template compatibility, we don't use related ourselves but this
            # method is only used in touchtechnology.content.admin
            "related": (),
        }
        context.update(extra_context)

        if templates is None:
            templates = self.template_path("edit.related.html", model._meta.model_name)

        return self.render(request, templates, context)

    @require_POST_m
    def generic_delete(
        self,
        request,
        model_or_manager,
        pk,
        post_delete_redirect=None,
        perms=None,
        permission_required=False,
        accept_global_perms=True,
        return_403=None,
        **kwargs,
    ):

        model, manager = model_and_manager(model_or_manager)

        instance = get_object_or_404(manager, pk=pk)

        # If not explicitly declared, redirect the user to login if they are
        # currently anonymous. If they are already identified, we can safely
        # return an HTTP 403 Forbidden.
        if return_403 is None:
            return_403 = not request.user.is_anonymous

        # Implement permission checking for specified object instance, throw
        # an HTTP 403 (using django-guardian configuration) when the user is
        # not entitled to edit the instance.
        if permission_required:

            # If no perms are specified, build sensible default using built in
            # permission types that come batteries included with Django.
            if perms is None:
                perms = get_perms_for_model(model, delete=True)

            # Determine the user's permission to edit this object using the
            # get_40x_or_None - saves decorating view method with
            # @permission_required_or_403
            has_permission = get_40x_or_None(
                request,
                perms,
                obj=instance,
                return_403=return_403,
                accept_global_perms=accept_global_perms,
            )

            # If permission is denied, return the response. May already have
            # thrown an exception by now.
            if has_permission is not None:
                return has_permission

        if post_delete_redirect is None:
            redirect_to = urljoin(request.path, "../..")
            post_delete_redirect = self.redirect(redirect_to)

        instance.delete()
        messages.add_message(
            request,
            messages.SUCCESS,
            _("The %s has been deleted.") % model._meta.verbose_name,
        )
        return post_delete_redirect

    def context_instance(self, request, **kwargs):
        return RequestContext(request, current_app=self.current_app, **kwargs)

    def render_to_string(self, request, templates, context, **kwargs):
        context_instance = self.context_instance(request)
        return render_to_string(templates, context, context_instance, **kwargs)

    def render(self, request, templates, context, tz=None, **kwargs):
        logger.debug(
            "{app}.render: {request.path}".format(
                app=self.__class__.__name__, request=request
            )
        )
        context.update({"application": self})
        kwargs.setdefault("context", context)

        response = TemplateResponse(request, templates, **kwargs)
        response.current_app = self.current_app

        if tz is not None:
            with timezone.override(tz):
                response.render()

        if not request.user.is_anonymous:
            patch_cache_control(response, private=True)
        return response


class AccountsSite(Application):
    def __init__(self, name="accounts", app_name="accounts", *args, **kwargs):
        super(AccountsSite, self).__init__(name=name, app_name=app_name)
        self.profile_form_class = kwargs.pop("profile_form_class", PROFILE_FORM_CLASS)
        self.user_model = get_user_model()

    @property
    def set_password_form(self):
        class SetPasswordForm(forms.SetPasswordForm):
            def __init__(slf, user, *a, **kw):
                super(SetPasswordForm, slf).__init__(user, *a, **kw)
                if user:
                    slf.user = self.user_model.objects.get(pk=user.pk)

        return SetPasswordForm

    @property
    def password_change_form(self):
        class PasswordChangeForm(forms.PasswordChangeForm):
            def __init__(slf, user, *a, **kw):
                super(PasswordChangeForm, slf).__init__(user, *a, **kw)
                if user:
                    slf.user = self.user_model.objects.get(pk=user.pk)

        return PasswordChangeForm

    @property
    def template_base(self):
        return "registration"

    def get_urls(self):
        urlpatterns = [
            path("login/", self.login, name="login"),
            path("logout/", self.logout, name="logout"),
            path("password-change/", self.password_change, name="password_change"),
            path(
                "password-change/done/",
                self.password_change_done,
                name="password_change_done",
            ),
            path("password-reset/", self.password_reset, name="password_reset"),
            path(
                "password-reset/done/",
                self.password_reset_done,
                name="password_reset_done",
            ),
            path(
                "reset/<uidb64>/<token>/",
                self.password_reset_confirm,
                name="password_reset_confirm",
            ),
            path(
                "reset/done/",
                self.password_reset_complete,
                name="password_reset_complete",
            ),
            path("profile/", self.profile, name="profile"),
            path(r"", self.index, name="index"),
        ]
        return urlpatterns

    @node2extracontext
    def login(self, request, extra_context, *args, **kwargs):
        form_class = getattr(
            settings,
            "AUTHENTICATION_FORM_CLASS",
            "django.contrib.auth.forms.AuthenticationForm",
        )
        authentication_form = import_string(form_class)
        view = views.LoginView.as_view(
            authentication_form=authentication_form, extra_context=extra_context
        )
        return view(request, *args, **kwargs)

    @node2extracontext
    def logout(self, request, extra_context, *args, **kwargs):
        view = views.LogoutView.as_view(extra_context=extra_context)
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_change(self, request, extra_context, *args, **kwargs):
        success_url = kwargs.pop(
            "post_change_redirect", self.reverse("password_change_done")
        )
        form_class = kwargs.pop("password_change_form", self.password_change_form)
        view = views.PasswordChangeView.as_view(
            form_class=form_class, success_url=success_url, extra_context=extra_context
        )
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_change_done(self, request, extra_context, *args, **kwargs):
        view = views.PasswordChangeDoneView.as_view(extra_context=extra_context)
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_reset(self, request, extra_context, *args, **kwargs):
        success_url = kwargs.pop(
            "post_reset_redirect", self.reverse("password_reset_done")
        )
        view = views.PasswordResetView.as_view(
            success_url=success_url, extra_context=extra_context
        )
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_reset_done(self, request, extra_context, *args, **kwargs):
        view = views.PasswordResetDoneView.as_view(extra_context=extra_context)
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_reset_confirm(self, request, extra_context, *args, **kwargs):
        success_url = kwargs.pop(
            "post_reset_redirect", self.reverse("password_reset_complete")
        )
        form_class = kwargs.pop("set_password_form", self.set_password_form)
        view = views.PasswordResetConfirmView.as_view(
            form_class=form_class,
            success_url=success_url,
            extra_context=extra_context,
        )
        return view(request, *args, **kwargs)

    @node2extracontext
    def password_reset_complete(self, request, extra_context, *args, **kwargs):
        view = views.PasswordResetCompleteView.as_view(extra_context=extra_context)
        return view(request, *args, **kwargs)

    def activation_complete(self, request, *args, **kwargs):
        context = {}
        context.update(kwargs)
        templates = self.template_path("activation_complete.html")
        return self.render(request, templates, context)

    def index(self, request, *args, **kwargs):
        return redirect(settings.LOGIN_REDIRECT_URL)

    @login_required_m
    def profile(self, request, *args, **kwargs):
        form_class = import_string(self.profile_form_class)

        if request.method == "POST":
            form = form_class(data=request.POST, instance=request.user)
            if form.is_valid():
                if form.has_changed():
                    form.save()
                    messages.add_message(
                        request, messages.SUCCESS, _("Your changes have been saved.")
                    )
                else:
                    messages.add_message(
                        request, messages.INFO, _("You did not make any changes.")
                    )
                return redirect(".")
        else:
            form = form_class(instance=request.user)

        context = {
            "form": form,
        }
        context.update(kwargs)

        templates = self.template_path("profile.html")
        return self.render(request, templates, context)



import uuid

from django.conf import settings
from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.forms.models import modelform_factory
from django.urls import include, path
from django.utils.translation import gettext_lazy as _

from touchtechnology.admin.base import AdminComponent
from touchtechnology.admin.forms import GroupEditForm, UserEditForm
from touchtechnology.common.decorators import staff_login_required_m

ANONYMOUS_USER_ID = getattr(settings, "ANONYMOUS_USER_ID", None)


class UsersGroups(AdminComponent):
    verbose_name = _("Users & Groups")

    user_order_by = ("last_name", "first_name")

    group_class = Group
    group_order_by = ("name",)

    unprotected = False

    def __init__(self, app, name="auth", app_name="auth"):
        super(UsersGroups, self).__init__(app, name, app_name)
        self.user_class = get_user_model()

    @property
    def template_base(self):
        return "touchtechnology/admin"

    def get_urls(self):
        user_patterns = (
            [
                path("", self.list_users, name="list"),
                path("add/", self.edit_user, name="add"),
                path("<int:pk>/", self.edit_user, name="edit"),
                path("<int:pk>/delete/", self.delete_user, name="delete"),
                path("<int:pk>/permission/", self.perms_user, name="perms"),
            ],
            self.app_name,
        )

        group_patterns = (
            [
                path("", self.list_groups, name="list"),
                path("add/", self.edit_group, name="add"),
                path("<int:pk>/", self.edit_group, name="edit"),
                path("<int:pk>/delete/", self.delete_group, name="delete"),
                path("<int:pk>/permission/", self.perms_group, name="perms"),
            ],
            self.app_name,
        )

        urlpatterns = [
            path("", self.index, name="index"),
            path("user/", include(user_patterns, namespace="users")),
            path("group/", include(group_patterns, namespace="groups")),
        ]
        return urlpatterns

    def dropdowns(self):
        dl = (
            (_("Users"), self.reverse("users:list"), "user"),
            (_("Groups"), self.reverse("groups:list"), "users"),
        )
        return dl

    @staff_login_required_m
    def index(self, request, *args, **kwargs):
        return self.redirect(self.reverse("users:list"))

    @staff_login_required_m
    def list_users(self, request, **extra_context):
        queryset = self.user_class.objects.exclude(pk=ANONYMOUS_USER_ID).order_by(
            *self.user_order_by
        )
        return self.generic_list(
            request,
            queryset,
            paginate_by=25,
            permission_required=True,
            search=("first_name", "last_name", "email"),
            extra_context=extra_context,
        )

    @staff_login_required_m
    def edit_user(self, request, pk=None, **extra_context):
        # when performing a creation we need to pass an instance with a random
        # password already set so that the user can initiate a password reset.
        instance = None

        if pk is None:
            instance = self.user_class()
            # generate a password that we will never retain in the clear and
            # would be very hard to brute force attack due to it's length.
            password = uuid.uuid4().hex

            # set the random password so that it's not possible to sign-in to
            # this account without first performing a password reset cycle.
            instance.set_password(password)

        # create a ModelForm using our base UserEditForm, but overload the
        # Meta.fields with value from custom UserModel.USER_EDIT_FORM_FIELDS
        # if it is present. Always exclude certain fields.
        fields = getattr(self.user_class, "USER_EDIT_FORM_FIELDS", "__all__")
        exclude = ("password", "last_login", "date_joined")
        form_class = modelform_factory(
            self.user_class, form=UserEditForm, fields=fields, exclude=exclude
        )

        return self.generic_edit(
            request,
            self.user_class,
            pk=pk,
            instance=instance,
            form_class=form_class,
            related=(),
            permission_required=True,
            post_save_redirect=self.redirect(self.reverse("users:list")),
            extra_context=extra_context,
        )

    @staff_login_required_m
    def delete_user(self, request, pk, **extra_context):
        return self.generic_delete(
            request, self.user_class, pk=pk, permission_required=True
        )

    @staff_login_required_m
    def perms_user(self, request, pk, **extra_context):
        return self.generic_permissions(
            request, self.user_class, pk=pk, **extra_context
        )

    @staff_login_required_m
    def list_groups(self, request, **extra_context):
        queryset = self.group_class.objects.order_by(*self.group_order_by)
        return self.generic_list(
            request,
            queryset,
            paginate_by=25,
            permission_required=True,
            extra_context=extra_context,
        )

    @staff_login_required_m
    def edit_group(self, request, pk=None, **extra_context):
        return self.generic_edit(
            request,
            self.group_class,
            pk=pk,
            form_class=GroupEditForm,
            related=(),
            permission_required=True,
            post_save_redirect=self.redirect(self.reverse("groups:list")),
            extra_context=extra_context,
        )

    @staff_login_required_m
    def delete_group(self, request, pk, **extra_context):
        return self.generic_delete(
            request, self.group_class, pk=pk, permission_required=True
        )

    @staff_login_required_m
    def perms_group(self, request, pk, **extra_context):
        return self.generic_permissions(
            request, self.group_class, pk=pk, **extra_context
        )



from django.db import migrations

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0038_match_live_stream"),
    ]

    operations = [
        migrations.AddField(
            model_name="season",
            name="live_stream",
            field=touchtechnology.common.db.models.BooleanField(default=False),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations, models

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("content", "0004_remove_placeholderkeywordargument"),
    ]

    operations = [
        migrations.AlterField(
            model_name="page",
            name="last_modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="pagecontent",
            name="last_modified",
            field=models.DateTimeField(auto_now=True),
        ),
        migrations.AlterField(
            model_name="redirect",
            name="active",
            field=touchtechnology.common.db.models.BooleanField(
                default=True, verbose_name="Active"
            ),
        ),
        migrations.AlterField(
            model_name="redirect",
            name="permanent",
            field=touchtechnology.common.db.models.BooleanField(
                default=False, verbose_name="Permanent"
            ),
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0018_resync_field_and_model_defs"),
    ]

    operations = [
        migrations.AlterField(
            model_name="competition",
            name="clubs",
            field=touchtechnology.common.db.models.ManyToManyField(
                related_name="competitions", to="competition.Club", blank=True
            ),
        ),
    ]



from django.test.utils import override_settings
from test_plus import TestCase


@override_settings(ROOT_URLCONF="example_app.urls")
class PaginationTest(TestCase):

    # Reuse the existing fixture, but might be better to use factory_boy to
    # create and arbitrary set of values in case the query_string fixture ever
    # needs to change.
    fixtures = ["query_string"]

    def test_setting_only(self):
        self.get("pagination:setting_only")
        n = len(self.get_context("object_list"))
        self.assertEqual(n, 5)

    def test_parameter(self):
        self.get("pagination:parameter", 3)
        n = len(self.get_context("object_list"))
        self.assertNotEqual(n, 5)
        self.assertEqual(n, 3)

    def test_parameter_no_pagination(self):
        self.get("pagination:parameter", 0)
        n = len(self.get_context("object_list"))
        self.assertIsNone(self.get_context("paginator"))
        self.assertEqual(n, 12)  # if fixture changes, may need updating



# -*- coding: utf-8 -*-

from django.db import migrations
from django.utils import timezone

import touchtechnology.common.db.models


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0001_initial"),
    ]

    operations = [
        migrations.AddField(
            model_name="sitemapnode",
            name="last_modified",
            field=touchtechnology.common.db.models.DateTimeField(
                default=timezone.now(), auto_now=True
            ),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("common", "0002_sitemapnode_last_modified"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PythonPackage",
        ),
        migrations.AlterModelOptions(
            name="sitemapnode",
            options={"verbose_name": "folder", "ordering": ("tree_id", "lft")},
        ),
    ]



# -*- coding: utf-8 -*-

import uuid

import django.db.models.deletion
from django.db import migrations, models

import touchtechnology.common.db.models


def initial_uuid(apps, schema_editor):
    Person = apps.get_model("competition", "Person")
    for p in Person.objects.select_for_update():
        p.uuid = uuid.uuid4()
        p.save()


class Migration(migrations.Migration):

    dependencies = [
        ("competition", "0019_auto_20161210_0922"),
    ]

    operations = [
        # Add the new uuid field to the Person model with null=True (will
        # breach NOT NULL constraints on creation if we don't do this)
        migrations.AddField(
            model_name="person",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                null=True,
            ),
        ),
        # Apply a data migration step to populate each existing instance with a
        # suitable uuid value
        migrations.RunPython(initial_uuid, reverse_code=migrations.RunPython.noop),
        # Now that all extsing records have a uuid value, alter the schema to
        # enforce the NOT NULL constraint moving forward
        migrations.AlterField(
            model_name="person",
            name="uuid",
            field=models.UUIDField(
                default=uuid.uuid4,
                editable=False,
                unique=True,
            ),
        ),
        # Each of the models which have a ForeignKey to the Person model we
        # need to add a new field which we can populate in a data migration;
        # each field needs to be NULL-able and must specify the "to_field"
        # attribute to make sure we point at the right field on the Person
        # model
        migrations.AddField(
            model_name="club",
            name="primary_uuid",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                to_field="uuid",
                on_delete=django.db.models.deletion.SET_NULL,
                related_name="+",
                blank=True,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="clubassociation",
            name="person_uuid",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                to_field="uuid",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                blank=True,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="seasonassociation",
            name="person_uuid",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                to_field="uuid",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                blank=True,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="teamassociation",
            name="person_uuid",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                to_field="uuid",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                blank=True,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="simplescorematchstatistic",
            name="player_uuid",
            field=touchtechnology.common.db.models.ForeignKey(
                to="competition.Person",
                to_field="uuid",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                blank=True,
                null=True,
            ),
        ),
    ]



import logging

logger = logging.getLogger(__name__)


def strict_show_toolbar_callback(request):
    """
    Additional check against custom permission ``debug_toolbar.show``.

    :param request: HttpRequest
    :return: bool
    """
    if hasattr(request, "user") and request.user.has_perm("debug_toolbar.show"):

        logger.debug(
            'DjDT allowed for "%s" <%s>',
            request.user.get_full_name(),
            request.user.email,
        )

        return show_toolbar_callback(request)

    return False


def show_toolbar_callback(request):
    """
    For each request, the callback will be invoked to determine if the
    machinery of the django-debug-toolbar should be invoked.

    In consideration of performance, this should be entirely stateless and
    depend on information available in the request only.

    We allow two ways that a user can enable this.

    1. User-Agent spoofing

    Setting the user agent header to include the string "DjangoDebugToolbar"
    will enable the toolbar for that request.

    2. X-DJDT-SHOW header

    Adding a custom header and setting the value to one of "1", "true", "on",
    or "yes" (all case insensitive) will enable the toolbar for that request.

    NOTE: previously this depended on a custom permission, we've relaxed this
    because sometimes it is necessary to debug a live application as an
    unauthenticated user.

    For the old behaviour, use the ``strict_show_toolbar_callback`` instead.

    :param request: HttpRequest
    :return: bool
    """
    ua = request.META.get("HTTP_USER_AGENT", "")
    if ua.find("DjangoDebugToolbar") >= 0:
        logger.info('Enable DjDT by User-Agent "%s"', ua)
        return True

    show = request.META.get("HTTP_X_DJDT_SHOW")
    if show is not None:
        show = show.lower()
        logger.debug("HTTP_X_DJDT_SHOW: %s", show)
    if show in ("1", "true", "on", "yes"):
        logger.info('Enable DjDT by X-DjDT-SHOW header "%s"', show)
        return True

    return False



import decimal
import operator

from pyparsing import (
    Forward,
    Literal,
    Optional,
    StringEnd,
    Word,
    ZeroOrMore,
    alphanums,
    alphas,
    nums,
)

__all__ = ("Calculator", "BonusPointCalculator")

Zero = decimal.Decimal(0)

# Atoms

Identifier = Word(alphas, alphanums + "_").setName("identifier")
Integer = (
    (Optional("-") + Word(nums))
    .setParseAction(lambda s, loc, toks: "".join(toks))
    .setName("integer")
)

# Literals

LeftParenthesis = Literal("(")
RightParenthesis = Literal(")")

LeftBracket = Literal("[")
RightBracket = Literal("]")

Colon = Literal(":")
Comma = Literal(",")

# Groups of Literals (all OR cases)
# Note: order matters, longer tokens must be listed first!
#   eg. '>=' before '>'

Addition = (Literal("+") | Literal("-")).setName("addop")
Multiplication = (Literal("*") | Literal("/")).setName("mulop")
Comparison = (
    Literal("=") | Literal(">=") | Literal("<=") | Literal(">") | Literal("<")
).setName("comparison")


class Calculator(object):

    def __init__(self, instance, *args, **kwargs):
        Expression = Forward()
        Atom = (
            (Identifier | Integer).setParseAction(self._push)
            | (LeftParenthesis + Expression.suppress() + RightParenthesis)
        ).setName("atom")
        Terminal = Atom + ZeroOrMore((Multiplication + Atom).setParseAction(self._push))
        Expression << Terminal + ZeroOrMore(
            (Addition + Terminal).setParseAction(self._push)
        )

        self.Operators = {
            "+": operator.add,
            "-": operator.sub,
            "*": operator.mul,
            "/": operator.truediv,
        }

        self.instance = instance
        self.pattern = Expression + StringEnd()
        self.stack = []

    def evaluate(self):
        try:
            token = self.stack.pop()
        except IndexError:
            return None
        if token in self.Operators:
            func = self.Operators.get(token)
            rhs = self.evaluate()
            lhs = self.evaluate()
            return func(lhs, rhs)
        elif Identifier.re.match(token):
            return getattr(self.instance, token, Zero)
        try:
            return decimal.Decimal(token)
        except decimal.InvalidOperation:
            return Zero

    def parse(self, data):
        self.pattern.parseString(data)

    def _push(self, string, position, tokens):
        """Add an element to the parse stack"""
        self.stack.append(tokens[0])


class BonusPointCalculator(Calculator):
    """
    Examples:

        Sydney Uni Touch
        "[win=1, score_against=0, forfeit_for=0: 1] + [loss=1, margin<=2: 1]"

        Super 14 Rugby
        "[tries>=4: 1] + [loss=1, margin<=7: 1]"
    """

    def __init__(self, *args, **kwargs):
        super(BonusPointCalculator, self).__init__(*args, **kwargs)

        self.Operators.update(
            {
                "=": operator.eq,
                ">": operator.gt,
                "<": operator.lt,
                ">=": operator.ge,
                "<=": operator.le,
                ",": operator.and_,
                "[": lambda a, b: a and b or Zero,
                "]": lambda a, b: a and b or Zero,
            }
        )

        BonusPointAddition = Addition | Comparison

        Expression = Forward()
        Atom = (Identifier | Integer).setParseAction(self._push) | (
            LeftParenthesis + Expression.suppress() + RightParenthesis
        )
        Terminal = Atom + ZeroOrMore((Multiplication + Atom).setParseAction(self._push))
        Expression << Terminal + ZeroOrMore(
            (BonusPointAddition + Terminal).setParseAction(self._push)
        )
        Condition = Expression + ZeroOrMore(
            (Comma + Expression).setParseAction(self._push)
        )

        Rule = (
            LeftBracket + Condition + Colon + Expression + RightBracket
        ).setParseAction(self._push)
        Chain = Rule + ZeroOrMore((Addition + Rule).setParseAction(self._push))

        self.pattern = Chain + StringEnd()

    def parse(self, data):
        if data.strip():
            self.pattern.parseString(data)



import logging
import subprocess

import requests
from django.conf import settings
from django.utils.html import strip_spaces_between_tags

logger = logging.getLogger(__name__)


def prince(html, base_url=None, ttl=300, **kwargs):
    SERVER = getattr(settings, "PRINCE_SERVER", None)
    BINARY = getattr(settings, "PRINCE_BINARY", "/usr/local/bin/prince")

    if base_url is None:
        base_url = getattr(settings, "PRINCE_BASE_URL", None)

    # When celery and django-tenant-schemas are involved, this get's a bit
    # weird. This has bitten once in production so lets log it and see else
    # might cause grief.
    for kw, arg in kwargs.items():
        logger.warning("Unexpected keyword argument: %s=%r", kw, arg)

    logger.debug("base_url: %s", base_url)
    logger.debug("ttl: %s", ttl)

    logger.debug("content-length: %s", len(html))
    html = strip_spaces_between_tags(html)
    logger.debug("content-length-stripped: %s", len(html))

    if SERVER:
        logger.debug("server: %s", SERVER)
        html = html.encode("utf8")

        headers = {}
        if base_url:
            headers["base_url"] = base_url
        for key, value in headers.items():
            logger.debug("header: %s=%s", key, value)

        res = requests.post(f"https://{SERVER}/", data=html, headers=headers)
        res.raise_for_status()

        out = res.content
        logger.info("Remote PDF generation via %s complete.", SERVER)

    else:
        command = [BINARY, "--input=html"]
        if base_url:
            command += ["--baseurl=%s" % base_url]
        command += ["-"]
        logger.debug("command: %s", " ".join(command))

        p = subprocess.Popen(command, stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        out, err = p.communicate(html.encode("utf8"))
        logger.info("Local PDF generation complete.")

        if err:
            logger.error(err)

    return out



from django.db import models

from tournamentcontrol.competition.query import (
    LadderEntryQuerySet,
    MatchQuerySet,
)


class LadderEntryManager(models.Manager.from_queryset(LadderEntryQuerySet)):
    def get_queryset(self):
        return super(LadderEntryManager, self).get_queryset()._all()


class MatchManager(models.Manager.from_queryset(MatchQuerySet)):
    def get_queryset(self):
        return super(MatchManager, self).get_queryset()._rank_importance()



from test_plus import TestCase

from tournamentcontrol.competition.admin import next_related_factory
from tournamentcontrol.competition.forms import TeamForm
from tournamentcontrol.competition.models import Team
from tournamentcontrol.competition.tests import factories


class TeamFormTests(TestCase):
    @classmethod
    def setUpTestData(cls):
        cls.clubs = factories.ClubFactory.create_batch(5)
        cls.division = factories.DivisionFactory.create()
        cls.division.season.competition.clubs.set(cls.clubs)
        # Prepare a new team instance that we can use for tests related to creating new teams.
        cls.new_team = next_related_factory(Team, cls.division)

    def test_empty_club_requires_title(self):
        "When the club is empty, a title is required."
        form = TeamForm(division=self.division, instance=self.new_team, data={})
        self.assertFormError(form, "title", ["You must specify a name for this team."])

    def test_populated_club_defaults_title(self):
        "When the club is populated, a title is not required."
        form = TeamForm(
            division=self.division,
            instance=self.new_team,
            data={"club": self.clubs[0].pk},
        )
        self.assertFormError(form, "title", [])
        self.assertQuerySetEqual(Team.objects.all(), [form.save()])



import itertools
from datetime import date, datetime
from zoneinfo import ZoneInfo

from test_plus import TestCase

from tournamentcontrol.competition.draw import (
    tournament_date_generator,
    weekly_date_generator,
)
from tournamentcontrol.competition.tests.factories import StageFactory


class DrawGenerationUtilities(TestCase):
    def setUp(self):
        super().setUp()
        self.stage = StageFactory.create(
            division__games_per_day=3,
            division__season__start_date=date(2022, 8, 2),
        )

    def test_tournament_date_generator(self):
        with self.assertLogs("tournamentcontrol.competition.draw", "DEBUG") as cm:
            res = tournament_date_generator(self.stage)
            self.assertCountEqual(
                itertools.islice(res, 10),
                [
                    date(2022, 8, 2),
                    date(2022, 8, 2),
                    date(2022, 8, 2),
                    date(2022, 8, 3),
                    date(2022, 8, 3),
                    date(2022, 8, 3),
                    date(2022, 8, 4),
                    date(2022, 8, 4),
                    date(2022, 8, 4),
                    date(2022, 8, 5),
                ],
            )

        self.assertCountEqual(
            cm.output,
            [
                f"DEBUG:tournamentcontrol.competition.draw:"
                f"Tournament date generator: '{self.stage.title}' (2022-08-02 00:00:00+00:00:3)",
            ],
        )

    def test_weekly_date_generator(self):
        with self.assertLogs("tournamentcontrol.competition.draw", "DEBUG") as cm:
            res = weekly_date_generator(self.stage)
            self.assertCountEqual(
                itertools.islice(res, 5),
                [
                    datetime(2022, 8, 2, tzinfo=ZoneInfo("UTC")),
                    datetime(2022, 8, 9, tzinfo=ZoneInfo("UTC")),
                    datetime(2022, 8, 16, tzinfo=ZoneInfo("UTC")),
                    datetime(2022, 8, 23, tzinfo=ZoneInfo("UTC")),
                    datetime(2022, 8, 30, tzinfo=ZoneInfo("UTC")),
                ],
            )

        self.assertCountEqual(
            cm.output,
            [
                f"DEBUG:tournamentcontrol.competition.draw:"
                f"Weekly date generator: '{self.stage.title}' (2022-08-02 00:00:00+00:00)",
            ],
        )



from django.apps import AppConfig


class NewsConfig(AppConfig):
    name = "touchtechnology.news"

    def ready(self):
        from touchtechnology.admin.sites import site
        from touchtechnology.news.admin import NewsAdminComponent

        site.register(NewsAdminComponent)
